//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  edif.cpp
//! \brief NEdif: inspect and edit data and metadata

#include "defs.hpp"

#include <boost/format.hpp>
#include <fstream>

#include "../readplus/ask.hpp"
#include "../trivia/file_ops.hpp"
#include "../trivia/string_ops.hpp"
#include "../trivia/vector_ops.hpp"

#include "edif.hpp"
#include "expr.hpp"
#include "fsel.hpp"
#include "genus.hpp"
#include "loop.hpp"
#include "mem.hpp"
#include "obj.hpp"
#include "olf.hpp"
#include "slice.hpp"
#include "variables.hpp"
#include "xax_lex.hpp"

using boost::format;


//**************************************************************************************************
//*  Show listings
//**************************************************************************************************

//! Display list of online files.

void NEdif::show_files()
{
    FileIterator fiter(triv::iota_list(SMem::instance()->mem_size()), true);
    while (const COlo* f = fiter.next()) {
        const COlc* fc = dynamic_cast<const COlc*>(f);
        const COld* fd = dynamic_cast<const COld*>(f);
        // prepare ztext
        string ztext;
        if (f->nZ()) {
            if (f->ZCo.size() == 1) {
                ztext = f->ZCo[0].str_compact(16);
            } else if (f->ZCo.size() == 2) {
                ztext = f->ZCo[0].str_compact(10) + ", ";
                ztext += f->ZCo[1].str_compact(16 - ztext.length());
            } else if (f->ZCo.size() == 3) {
                ztext = f->ZCo[0].str_compact(8) + ",";
                ztext += f->ZCo[1].str_compact(16 - ztext.length() - 4) + ",";
                ztext += f->ZCo[2].str_compact(16 - ztext.length());
            } else {
                ztext = f->ZCo[0].str_compact(6) + ",";
                ztext += f->ZCo[1].str_compact(16 - ztext.length() - 9) + ",";
                ztext += f->ZCo[2].str_compact(16 - ztext.length() - 6) + ",";
                ztext += f->ZCo[3].str_compact(16 - ztext.length() - 3) + ",..";
            }
        } else {
            ztext = string(16, ' ');
        }
        // prepare xytext
        string xytext;
        if (fd) {
            int npts = fd->nPts();
            string txtpts;
            if (npts)
                txtpts = str(format("%-6d") % npts);
            else
                txtpts = "     ?";
            xytext = txtpts + f->xco.str_compact(12) + " " + f->yco.str_compact(12);
        } else if (fc) {
            xytext = fc->expr.substr(0, 25);
        }
        // print everything
        cout << format("%3i%1s%-16s %5i %-16s %-32s\n") % fiter.k() % (f->as_on_disk ? "=" : " ")
                % f->name % f->nJ() % ztext % xytext;
    }
}


//! Display list of z-values and x-ranges for spectra.

void NEdif::show_spectra()
{
    FileIterator fiter(SFSel::instance()->sel());
    while (const COlo* f = fiter.next()) {
        cout << "file " << fiter.k() << ":\n";
        if (const COld* fd = dynamic_cast<const COld*>(f)) {
            printf("%3s ", "j");
            for (const CCoord& zco : f->ZCo)
                printf(" %12.12s", zco.str_compact().c_str());
            printf(" %12s (%4s) %12s\n", "x_inf", "#pts", "x_sup");
            for (int j = 0; j < f->nJ(); j++) {
                printf("%3i%c", j, f->V[j]->frozen ? '/' : ' ');
                for (RObj z : f->V[j]->z)
                    printf(" %s", z->to_s(12, 12, 6).c_str());
                printf(
                    " %12.6g (%4i) %12.6g\n", fd->VS(j)->x[0], fd->nPts(j),
                    fd->VS(j)->x[fd->nPts(j) - 1]);
            }
        } else if (const COlc* fc = dynamic_cast<const COlc*>(f)) {
            fc->print_info_table("");
        }
    }
}


//! Display coordinate names and units.

void NEdif::show_coord()
{
    FileIterator fiter(SFSel::instance()->sel());
    while (const COlo* f = fiter.next()) {
        string out = str(format("%3i") % fiter.k());
        out += " x: " + f->xco.str_compact();
        out += "  y: " + f->yco.str_compact();
        for (int i = 0; i < f->nZ(); i++)
            out += "  z" + S(i) + ": " + f->ZCo[i].str_compact();
        cout << out << "\n";
    }
}


//! Display z coordinates (names and units).

void NEdif::show_coordZ()
{
    FileIterator fiter(SFSel::instance()->sel());
    const COlo* f = fiter.next();
    // Get current coordinates:
    vector<CCoord> ZCo = f->ZCo;
    ;
    int nz = ZCo.size();
    if (nz)
        for (int i = 0; i < nz; ++i)
            printf("  %i: %s\n", i, ZCo[i].str_compact().c_str());
    else
        printf("  there are no z coordinates\n");

    // Check other input files:
    while (f = fiter.next()) {
        int k = fiter.k();
        if (f->ZCo.size() != nz)
            throw "file " + S(k) + " has different number of z coordinates";
        for (int i = 0; i < nz; ++i)
            if (f->ZCo[i] != ZCo[i])
                throw "file " + S(k) + " has different z coordinate " + S(i);
    }
}


//! Display numeric parameters.

void NEdif::show_numpars()
{
    FileIterator fiter(SFSel::instance()->sel());
    while (const COlo* f = fiter.next()) {
        cout << "file " << fiter.k() << ":\n";
        for (int m = 0; m < f->RPar.size(); m++)
            cout << "  r" << m
                 << ": " << f->RPar[m].Co.str_compact()
                 << ": " << f->RPar[m].val << "\n";
    }
}


//! Display documentation lines.

void NEdif::show_doc()
{
    FileIterator fiter(SFSel::instance()->sel());
    while (const COlo* f = fiter.next()) {
        if (fiter.size() > 1)
            cout << "# file " << fiter.k() << ":\n";
        for (int i = 0; i < f->lDoc.size(); i++)
            cout << f->lDoc[i] << "\n";
    }
}


//**************************************************************************************************
//*  Edit data
//**************************************************************************************************


//! Modify file name

void NEdif::edit_fnam()
{
    FileIterator fiter(SFSel::instance()->sel());
    while (COlo* f = fiter.next()) {
        string fnam = wask(("Rename " + f->name).c_str(), f->name);
        if (fnam != f->name) {
            f->log_action("ef " + fnam + " # < " + f->name);
            f->name = fnam;
        }
    }
}

//! Modify coordinate names and units.

void NEdif::edit_coord(string which)
{ // EMBEDDED_DIALOG
    CGenus genus(which);

    FileIterator fiter(SFSel::instance()->sel());
    // check whether all files have the same coordinate:
    COlo* fin = fiter.next();
    CCoord old_co = fin->coord(genus);
    bool uniform = true;
    bool multifile = false;
    while (fin = fiter.next()) {
        multifile = true;
        if (fin->coord(genus) != old_co) {
            uniform = false;
            break;
        }
    }

    // list current values:
    if (multifile && !uniform) {
        fiter.reset();
        while (fin = fiter.next()) {
            cout << "  file " << fiter.k() << ": " << fin->coord(genus) << "\n";
        }
    }
    string in = wask(
        "Coordinate " + which + ": name(unit)" + (multifile ? " or f to proceed file by file" : ""),
        uniform ? old_co.str_compact() : "f");
    fiter.reset();
    if (multifile && in == "f") {
        while (fin = fiter.next()) {
            old_co = fin->coord(genus);
            in = wask(
                "Coordinate " + which + " for file " + S(fiter.k()) + ": name(unit)",
                old_co.str_compact());
            CCoord new_co(in);
            fin->log_action(
                "ec" + string(which) + " " + new_co.str_compact() + " # old: "
                + old_co.str_compact());
            fin->set_coord(genus, new_co);
        }
    } else {
        CCoord new_co(in);
        while (fin = fiter.next()) {
            fin->log_action(
                "ec" + string(which) + " " + new_co.str_compact() + " # old: "
                + old_co.str_compact());
            fin->set_coord(genus, new_co);
        }
    }
}


//! Edit documentation lines.

void NEdif::edit_doc()
{
    FileIterator fiter(SFSel::instance()->sel());
    COlo* f;

    // obtain unused tmp file name:
    string fname = triv::next_tmp_file("/tmp/frida.%i");

    // transfer to tmp file:
    std::ofstream Fo(fname.c_str());
    while (f = fiter.next()) {
        Fo << "### FILE " << fiter.k() << " " << f->name << " ###\n";
        for (string lin : f->lDoc)
            Fo << lin << "\n";
    }
    Fo.close();

    // let user edit that file:
    system(("vi " + fname).c_str());

    // read back from tmp file:
    fiter.reset();
    FILE* Fi;
    if (!(Fi = fopen(fname.c_str(), "r")))
        throw S("cannot read back from tmp file");
    string line;
    while (triv::freadln(Fi, &line)) {
        if (line.substr(0, 3) == "###") {
            f = fiter.next();
            if (line.substr(3, 6) != " FILE ")
                throw S("invalid change in ### line: 'FILE' not found");
            string aux;
            triv::string_extract_word(line.substr(9), &aux, &line);
            int kin;
            if (!triv::any2int(aux, &kin))
                throw S("invalid change in ### line: file number not found");
            if (kin != fiter.k())
                throw S("invalid change in ### line: wrong file number");
            triv::string_extract_word(line, &aux, &line);
            if (aux != f->name)
                throw S("invalid change in ### line: wrong file name");
            f->lDoc.clear();
        } else {
            f->log_action(line);
        }
    }

    // delete tmp file:
    fclose(Fi);
    system(("rm -f " + fname).c_str());
}


//! Edit a variable. Preferred over a assignement if no return value shall be displayed.

void NEdif::edit_var()
{
    string name = wask("Create or modify which variable");
    string expr = wask("Value");
    RNode T = user_xaxparse(expr.c_str());
    RObj ret = T->tree_point_val();
    SVariRegistry::instance()->register_scalar(name, ret);
}

//! Delete numeric parameter.

void NEdif::del_numpar()
{
    int iN = zask("Index of numeric parameter to delete");
    FileIterator fiter(SFSel::instance()->sel());
    while (COlo* f = fiter.next()) {
        if (iN >= f->RPar.size())
            throw "file " + S(fiter.k()) + " has no numpar n" + S(iN);
        f->RPar.erase(f->RPar.begin() + iN);
    }
}
