//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  fit.cpp
//! \brief NCurveFit: fit a curve file to a data file

#include "defs.hpp"

#include <boost/format.hpp>
#include <future>
#include <lmmin.h>
#include <mutex>

#include "../readplus/ask.hpp"

#include "fit.hpp"
#include "fsel.hpp"
#include "loop.hpp"
#include "mem.hpp"
#include "expr.hpp"
#include "obj.hpp"
#include "olf.hpp"
#include "slice.hpp"

using boost::format;

bool allow_slow_conv = true;

namespace NCurveFit
{
string fit_one_spec(COlc* fc, const COld* fd, int k, int j, const lm_control_struct& control);
void fit_global(COlc* fc, const COld* fd, int k, const lm_control_struct& control);
}

std::mutex MX1, MX2;

void threadloop(
    const int niter, int* jTodo, const std::function<string(int)>& exec,
    vector<string>* results, int* jReported, int reportFactor)
{
    while(true) {
        MX1.lock();
        int j = (*jTodo)++;
        MX1.unlock();
        if( j>=niter )
            return; // nothing left to do, terminate thread
        (*results)[j] = exec(j);
        MX2.lock();
        for (int jj=*jReported+1; jj<niter; ++jj) {
            if ((*results)[jj]=="")
                break;
            if (!(jj%reportFactor))
                cout << (*results)[jj];
            *jReported = jj;
        }
        MX2.unlock();
    }
}

//! Returns a factor that reduces n output lines to a screen full.
//! Possible results are 1, 2, 4, 10, 20, ...
int computeReportingFactor(int n)
{
    if (n>=24) {
        int order = floor(log10(n));
        double logMantisse = log10(n)-order;
        if (logMantisse>log10(4.85))
            return round(4*pow(10,order-1));
        else if (logMantisse>log10(2.45))
            return round(2*pow(10,order-1));
        else
            return round(1*pow(10,order-1));
    } else
        return 1;
}

void conditionally_parallel_for_with_report(
    bool condition, int niter, const std::function<string(int)>& exec,
    const std::function<void(const string&)>& report)
{
    if (condition) {
        int nThreads = std::thread::hardware_concurrency();
        if (!nThreads) {
            cout << "CANNOT OBTAIN NUMBER OF CORES, WON'T GO PARALLEL\n";
            nThreads = 1;
        } else if (nThreads>niter)
            nThreads = niter;
        int jTodo = 0, jReported = -1;
        int reportFactor = computeReportingFactor(niter);
        vector<string> results(niter);
        vector<std::thread> threads(nThreads);
        for (std::thread& t: threads)
            t = std::thread(threadloop, niter, &jTodo, exec, &results, &jReported, reportFactor);
        for (std::thread& t: threads)
            t.join();
    } else {
        for (int i = 0; i < niter; ++i)
            report(exec(i));
    }
}

//**************************************************************************************************
//*  Fit tuning parameters
//**************************************************************************************************


namespace NCurveFit
{
double Epsilon = 1.e-14;
double Factor = 100.;
int nCall = 200;
int ScaleDiag = 1;
int verbosity = 0;
double maxtime = 3;
double FTol = 7e-15;
double GTol = 7e-15;
double XTol = 7e-15;
}


//! Set auxiliary parameters for the fit. Directly called from main program.

void NCurveFit::set_fit_tuning_pars(string which)
{
    if (which == "?") {
        cout << "Fit setup:\n";
        cout << "  [csc]  nCall = " << nCall << "\n";
        cout << "  [csdf] ftol = " << FTol << "\n";
        cout << "  [csdg] gtol = " << GTol << "\n";
        cout << "  [csdx] xtol = " << XTol << "\n";
        cout << "  [cse]  Differentiation step epsilon = " << Epsilon << "\n";
        cout << "  [csf]  Step bound factor = " << Factor << "\n";
        cout << "  [css]  Scale Diagonal = " << ScaleDiag << "\n";
        cout << "  [cst]  Timeout = " << maxtime << "s\n";
        cout << "  [csv]  Verbosity = " << verbosity << "\n";
    } else if (which == "v")
        verbosity = iask("Verbosity (between 0 and 15, or'ed flags logic)", verbosity);
    else if (which == "s")
        ScaleDiag = bask("Scale diagonal", ScaleDiag);
    else if (which == "c")
        nCall = iask("Maximum number of Levenberg-Marquardt steps", nCall);
    else if (which == "df")
        FTol = dask("ftol (relative error in the sum of squares)", FTol );
    else if (which == "dg")
        GTol = dask("gtol (orthogonality desired between fvec and its derivatives)", GTol );
    else if (which == "dx")
        XTol = dask("xtol (relative error between last two approximations)", XTol );
    else if (which == "f")
        Factor = dask("Step bound factor (generally between 0.01 and 100)", Factor);
    else if (which == "e")
        Epsilon = dask("Differentiation step epsilon", Epsilon);
    else if (which == "t")
        maxtime = dask("Max duration of fits (in s)", maxtime);
}


//**************************************************************************************************
//*  Common subroutines for different fit modes
//**************************************************************************************************

//! Returns zero when residues were computed sucessfully.

static int compute_residues(
    double* fvec, const int mdat, const int j, const CSpec* s, RObjVec& cu, const COlc::TWgt wt)
{
    try {
        int n = s->size();

        if ((wt == COlc::_VAR || wt == COlc::_VARD) && s->dy.size() != n)
            throw S("missing data error bars for weighing");

        for (int i = 0; i < n; i++) {

            if (wt == COlc::_LIN) {
                fvec[mdat + i] = s->y[i] - cu->to_r(i);
            } else if (wt == COlc::_LOG) {
                if (s->y[i] < 1e-99) {
                    throw "cannot weigh with log y: spectrum " + S(j) + " channel " + S(i)
                        + " has y=" + S(s->y[i]);
                } else if (cu->to_r(i) < 1e-99) {
                    fvec[mdat + i] = 99; // giant penalty
                } else {
                    fvec[mdat + i] = log(s->y[i] / cu->to_r(i));
                }
            } else if (wt == COlc::_VAR || wt == COlc::_VARC || wt == COlc::_VARD) {
                double dr;
                if (wt == COlc::_VAR)
                    dr = sqrt(SQR(s->dy[i]) + SQR(cu->to_dr(i)));
                else if (wt == COlc::_VARC)
                    dr = cu->to_dr(i);
                else if (wt == COlc::_VARD)
                    dr = s->dy[i];
                else
                    throw S("BUG: unforeseen case");
                if (dr < 1e-99)
                    throw "cannot weigh with dy: spectrum " + S(j) + " channel " + S(i)
                        + " has y{len=" + S(s->y.size()) + "}=" + S(s->y[i]) + ", dy{len="
                        + S(s->dy.size()) + "}=" + S(s->dy[i]) + "; theory yields y="
                        + S(cu->to_r(i)) + ", dy=" + (cu->has_err() ? S(cu->to_dr(i)) : S("n/a"));
                fvec[mdat + i] = (s->y[i] - cu->to_r(i)) / dr;
            }
        }
        return 0;
    } catch (string& s) {
        throw "cannot compute residues: "+s;
    } catch (...) {
        throw S("BUG: invalid exception in compute_residues\n");
    }
    return 1;
}


//**************************************************************************************************
//*  Least-squares fitting of single spectra (no global parameters)
//**************************************************************************************************

//! Data transfer between frida and lm_minimize callback functions.

typedef struct {
    const COld* fd;
    COlc* fc;
    int k;
    int j;
    double timeout;
} FitDatTyp;


//! Callback routine for use in lm_minimize: function evaluation.

static void
fit_evaluate(const double* par, int m_dat, const void* data, double* fvec, int* userbreak)
// arguments (required by lmmin):
//    par       r/w  function parameters, to be optimized
//    m_dat     r    number of data points
//    data      r    pointer to arbitrary data, here defined as FitDatTyp
//    fvec      w    vector of residues, to be minimized
//    userbreak w    can be use to terminate the minimization
{
    FitDatTyp* mydata {(FitDatTyp*)data};
    COlc* fc = mydata->fc;
    CCurve* c = fc->VC(mydata->j);
    try {
        if (mydata->timeout && time(nullptr) > mydata->timeout)
            throw S("reached fit time out");
        const COld* fd = mydata->fd;
        COlc::TWgt wt = fc->weighing;

        int ip = 0;
        for (int i = 0; i < fc->nP; ++i)
            if (fc->VC(mydata->j)->ParAttr[i] != 'x')
                c->P[i] = par[ip++];

        const CSpec* s = fd->VS(mydata->j);
        bool want_error = wt == COlc::_VAR || wt == COlc::_VARC;

        RObjVec cu = fc->eval_curve(s->x, mydata->k, mydata->j, want_error);

        if (compute_residues(fvec, 0, mydata->j, s, cu, wt))
            *userbreak = -1;

    } catch (string& s) {
        cout << "fit_evaluate j=" << mydata->j;
        for (int i = 0; i < fc->nP; ++i)
            cout << " p" << i << "=" << c->P[i];
        cout << ": " << s << "\n";
        *userbreak = -1;
    } catch (...) {
        cout << "BUG: fit_evaluate catched invalid exception\n";
        *userbreak = -1;
    }
}


//! Fits one spectrum. Used when there are no global parameters. Returns an info line.

string
NCurveFit::fit_one_spec(COlc* fc, const COld* fd, int k, int j, const lm_control_struct& control)
{
    try {
        CSpec* D = fd->VS(j);
        CCurve* C = fc->VC(j);
        if (C->frozen)
            throw S("frozen, no fit");
        FitDatTyp data;
        data.fc = fc;
        data.fd = fd;
        data.k = k;
        data.j = j;
        int nd = D->size();
        if (!nd)
            throw "empty data set in spectrum " + S(j);

        // Transfer parameters from C->P to Par
        //   (lmmin modifies Par; fit_evaluate transfers values back from Par to C->P)
        int np = fc->nP;
        vector<double> Par;
        for (int ip = 0; ip < np; ++ip)
            if (C->ParAttr[ip] != 'x')
                Par.push_back(C->P[ip]);
        int npfree = Par.size();
        vector<double> ParCovar(npfree*npfree);

        // Levenberg-Marquardt routine from library lmfit:

        data.timeout = time(nullptr) + std::thread::hardware_concurrency() * maxtime;

        lm_status_struct status;
        lmmin2(npfree, &(Par[0]), NULL, &(ParCovar[0]), nd, NULL, &data,
               fit_evaluate, &control, &status);
        if (status.outcome == 11)
            throw S("exception in fit function");

        for (int i = 0; i < np*np; ++i)
            C->PCovar[i] = 0;
        int ipfit = 0;
        for (int ip = 0; ip < np; ++ip) {
            if (C->ParAttr[ip] != 'x') {
                C->P[ip] = Par[ipfit];
                int jpfit = 0;
                for (int jp = 0; jp < np; ++jp)
                    if (C->ParAttr[jp] != 'x')
                        C->PCovar[ip*np+jp] = ParCovar[ipfit*npfree+(jpfit++)];
                ipfit++;
            }
        }

        // Fit quality metrics:
        C->fitOutcome = (double)status.outcome;
        C->nFunctionEvaluations = status.nfev;

        // coefficient of determination R^2 (for linear regression)
        double ysum1 = 0, ysum2 = 0;
        for (int i = 0; i < nd; ++i) {
            ysum1 += D->y[i];
            ysum2 += SQR(D->y[i]);
        }
        double ydev2 = ysum2 - SQR(ysum1) / nd; // = nd * <(y-<y>)^2>
        C->fitR2 = ydev2 != 0 ? 1 - SQR(status.fnorm) / ydev2 : -1;

        // mean deviation (chi^2 if fvec is weighed with 1/dy)
        int nfreedom = nd - npfree - 1;
        if (nfreedom < 1)
            nfreedom = 1;
        C->fitChi2 = SQR(status.fnorm) / nfreedom;

        string out = str(format("%2i %3i") % k % j);
        for (int ip = 0; ip < npfree; ip++)
            out += str(format(" %12g +- %12g") % Par[ip] % sqrt(ParCovar[ip*npfree+ip]));
        out += str(format(" (%3i->%9.7g)") % status.nfev % C->fitChi2);
        out += lm_shortmsg[status.outcome];
        return out;
    } catch (string& e) {
        return str(format("%3i%3i failed: %s") % k % j % e);
    } catch (...) {
        return str(format("%3i%3i failed with unexpected exception type") % k % j);
    }
}


//**************************************************************************************************
//*  Least-squares fitting of all spectra in a file (with global parameters)
//**************************************************************************************************

//! Data transfer between frida and lm_minimize callback functions.

typedef struct {
    const COld* fd;
    COlc* fc;
    vector<int> J2J;
    vector<int> Offset;
    int k;
    int npfree;
    double timeout;
} FitGloDatTyp;


//! Callback routine for use in lm_minimize: function evaluation.

static void
fit_evaluate_glo(const double* par, int m_dat, const void* data, double* fvec, int* userbreak)
// arguments (required by lmmin):
//    par       r/w  function parameters, to be optimized
//    m_dat     r    number of data points
//    data      r    pointer to arbitrary data, here defined as FitDatTyp
//    fvec      w    vector of residues, to be minimized
//    userbreak w    can be use to terminate the minimization
{
    try {
        FitGloDatTyp* mydata {(FitGloDatTyp*)data};
        if (mydata->timeout && time(nullptr) > mydata->timeout)
            throw S("reached fit time out");
        COlc* fc = mydata->fc;
        const COld* fd = mydata->fd;
        COlc::TWgt wt = fc->weighing;

        int npar = 0;
        for (int ip = 0; ip < fc->nP; ++ip) {
            if (fc->VC(mydata->J2J[0])->ParAttr[ip] == 'g') {
                for (int j : mydata->J2J)
                    fc->VC(j)->P[ip] = par[npar];
                npar++;
            } else {
                for (int j : mydata->J2J)
                    if (fc->VC(j)->ParAttr[ip] == 'u')
                        fc->VC(j)->P[ip] = par[npar++];
            }
        }
        if (npar != mydata->npfree)
            throw S("inconsistent number of free parameters");

        bool want_error = wt == COlc::_VAR || wt == COlc::_VARC;
        if (!NCurveFit::verbosity && mydata->J2J.size() > 1) {
            vector<std::future<int>> workers;
            for (int jj = 0; jj < mydata->J2J.size(); ++jj) {
                workers.push_back(std::async(
                    std::launch::async,
                    [&](int _jj) -> int {
                        RObjVec cu = fc->eval_curve(
                            fd->VS(mydata->J2J[_jj])->x, mydata->k, mydata->J2J[_jj], want_error);
                        return compute_residues(
                            fvec, mydata->Offset[_jj], mydata->J2J[_jj], fd->VS(mydata->J2J[_jj]),
                            cu, wt);
                    },
                    jj));
            }
            for (auto& w : workers) {
                if (w.get()) {
                    *userbreak = -1;
                    return;
                }
            }
        } else {
            for (int jj = 0; jj < mydata->J2J.size(); ++jj) {
                RObjVec cu = fc->eval_curve(
                    fd->VS(mydata->J2J[jj])->x, mydata->k, mydata->J2J[jj], want_error);
                if (compute_residues(
                        fvec, mydata->Offset[jj], mydata->J2J[jj], fd->VS(mydata->J2J[jj]), cu,
                        wt)) {
                    *userbreak = -1;
                    return;
                }
            }
        }
    } catch (string& s) {
        cout << "\n" << s << "\n";
        *userbreak = -1;
    } catch (const char* s) {
        cout << "\n" << s << "\n";
        *userbreak = -1;
    } catch (...) {
        cout << "BUG: fitEvaluate catched invalid exception\n";
        *userbreak = -1;
    }
}


//! Fits all spectra simultaneously. Used when there are global parameters.

void NCurveFit::fit_global(COlc* fc, const COld* fd, int k, const lm_control_struct& control)
{
    FitGloDatTyp data;
    data.fc = fc;
    data.fd = fd;
    data.k = k;

    // Set vector J2J to contain indices of non-frozen curves.
    for (int j = 0; j < fc->nJ(); ++j)
        if (!fc->VC(j)->frozen)
            data.J2J.push_back(j);
    if (!data.J2J.size())
        throw S("all curves are frozen");
    int njj = data.J2J.size();

    // Set vector Par to contain all free fit parameters.
    int np = fc->nP;
    vector<double> Par;
    for (int ip = 0; ip < np; ++ip) {
        int j0 = data.J2J[0];
        for (int jj = 1; jj < njj; ++jj) {
            int j = data.J2J[jj];
            if ((fc->VC(j0)->ParAttr[ip] == 'g') ^ (fc->VC(j)->ParAttr[ip] == 'g'))
                throw S("Global attribute must be uniform for all non-frozen spectra");
        }
        if (fc->VC(j0)->ParAttr[ip] == 'g') {
            Par.push_back(fc->VC(j0)->P[ip]);
        } else {
            for (int j : data.J2J)
                if (fc->VC(j)->ParAttr[ip] == 'u')
                    Par.push_back(fc->VC(j)->P[ip]);
        }
    }
    int npfree = Par.size();
    if (npfree <= 0)
        return;
    data.npfree = npfree;

    int nd = 0;
    for (int j : data.J2J) {
        data.Offset.push_back(nd);
        nd += fd->VS(j)->size();
    }

    // Levenberg-Marquardt routine from library lmfit:

    data.timeout = time(nullptr) + fc->nJ() * maxtime;

    lm_status_struct status;
    lmmin(npfree, &(Par[0]), nd, NULL, &data, fit_evaluate_glo, &control, &status);

    if (status.outcome == 11)
        throw S("exception in fit function");

    COlc::TWgt wt = fc->weighing;
    bool want_error = wt == COlc::_VAR || wt == COlc::_VARC;

    int nfreedom = std::max(1, nd - npfree - 1);
    fc->chi2 = SQR(status.fnorm) / nfreedom;
    cout << str(format("%3i global fit result: chi^2=%14.7g\n") % k % fc->chi2);

    for (int j : data.J2J) {
        fc->VC(j)->fitOutcome = (double)status.outcome;

        // compute deviation per spectrum:
        const CSpec* s = fd->VS(j);
        RObjVec cu = fc->eval_curve(s->x, k, j, want_error);
        int n = s->size();
        double* fvec = (double*)malloc(n * sizeof(double));
        compute_residues(fvec, 0, j, s, cu, wt);
        double fnorm = lm_enorm(n, fvec);
        free(fvec);

        // coefficient of determination R^2 (for linear regression)
        double ysum1 = 0, ysum2 = 0;
        for (int i = 0; i < n; ++i) {
            double y = fd->VS(j)->y[i];
            ysum1 += y;
            ysum2 += SQR(y);
        }
        double ydev2 = ysum2 - SQR(ysum1) / n;
        fc->VC(j)->fitR2 = ydev2 != 0 ? 1 - SQR(fnorm) / ydev2 : -1;

        // mean deviation (chi^2 if fvec is weighed with 1/dy)
        double xpfree = 0;
        for (int ip = 0; ip < np; ++ip) {
            if (fc->VC(j)->ParAttr[ip] == 'u')
                xpfree += 1;
            else if (fc->VC(j)->ParAttr[ip] == 'g')
                xpfree += 1.0 / fc->nJ();
        }
        double xfreedom = std::max(1.0, fd->VS(j)->size() - xpfree - 1);
        fc->VC(j)->fitChi2 = SQR(fnorm) / xfreedom;
    }

    // print results:
    for (int jj = 0; jj < njj; ++jj) {
        int j = data.J2J[jj];
        printf("%3i%3i", k, j);
        for (int ip = 0; ip < np; ip++) {
            if (fc->VC(j)->ParAttr[ip] == 'x')
                continue;
            printf(" %12g", fc->VC(j)->P[ip]);
            if (fc->VC(j)->ParAttr[ip] != 'u')
                printf("%c", fc->VC(j)->ParAttr[ip]);
        }
        printf(" (%3i->%22.16g)", status.nfev, fc->VC(j)->fitChi2);
        printf("%s\n", lm_shortmsg[status.outcome]);
    }
}


//**************************************************************************************************
//*  Least-squares fitting: main
//**************************************************************************************************

//! Fits selected curve files to the data files they refer to.

void NCurveFit::fit(bool _allow_slow_conv)
{
    lm_control_struct control = lm_control_double;
    allow_slow_conv = _allow_slow_conv;

    FileIterator fiter(SFSel::instance()->selC());

    while (COlc* fc = fiter.nextC()) {

        // assert validity of parameter attributes
        for (int j = 0; j < fc->nJ(); ++j)
            fc->VC(j)->assert_attr();

        COld* fd = SMem::instance()->mem_get_D(fc->kd);
        if (fc->nJ() != fd->nJ())
            throw S("data file and curve file have different number of spectra");
        if (!fc->nJ())
            throw S("BUG: empty curve and data file");
        if ((fc->weighing == COlc::_VAR || fc->weighing == COlc::_VARD) && !fd->VS(0)->has_dy())
            throw S("Data have no error bars, cannot apply chosen fit weights");

        control.epsilon = Epsilon;
        control.stepbound = Factor;
        control.patience = nCall;
        control.verbosity = verbosity;
        control.scale_diag = ScaleDiag;
        control.ftol = FTol;
        control.gtol = GTol;
        control.xtol = XTol;

        if (fc->has_global_par()) {
            fit_global(fc, fd, fiter.k(), control);
        } else {
            fc->chi2 = 0;
            vector<int> J;
            for (int j = 0; j < fc->nJ(); j++)
                if (!fc->VC(j)->frozen)
                    J.push_back(j);
            conditionally_parallel_for_with_report(
                !NCurveFit::verbosity, J.size(),
                [&](int _jj) -> string {
                    return fit_one_spec(fc, fd, fiter.k(), J[_jj], control) + "\n";
                },
                [&](const string& msg) -> void { cout << msg; });
        } // fit mode
    } // k
}
