//**************************************************************************************************
//* FRIDA: fast reliable interactive data analysis
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-
//* http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  node.cpp
//! \brief Nodes within expression trees, inheriting from virtual base CNode.

#include "defs.hpp"

#include <algorithm>

#include "curve.hpp"
#include "expr.hpp"
#include "fregistry.hpp"
#include "func.hpp"
#include "mem.hpp"
#include "node.hpp"
#include "obj.hpp"
#include "olf.hpp"
#include "reduce_curv.hpp"
#include "reduce_spec.hpp"
#include "slice.hpp"
#include "variables.hpp"

extern bool allow_slow_conv; // set in curve.cpp

const int debug = 0;

//**************************************************************************************************
//* CNodeWithArgs
//**************************************************************************************************

bool CNodeWithArgs::has_dummy() const
{
    bool ret = std::any_of(arg.begin(), arg.end(), [](RNode a) { return a && a->has_dummy(); });
    if (debug)
        std::cerr << ("CNodeWithArgs{" + tree_info() + "}.has_dummy -> " + S(ret) + "\n");
    return ret;
}

bool CNodeWithArgs::has_conv() const
{
    bool ret = std::any_of(arg.begin(), arg.end(), [](RNode a) { return a && a->has_conv(); });
    if (debug)
        std::cerr << ("CNodeWithArgs{" + tree_info() + "}.has_conv -> " + S(ret) + "\n");
    return ret;
}

int CNodeWithArgs::npar() const
{
    int ret = 0;
    for (RNode a : arg)
        if (a)
            ret = std::max(ret, a->npar());
    if (debug)
        std::cerr << ("CNodeWithArgs{" + tree_info() + "}.npar -> " + S(ret) + "\n");
    return ret;
}

bool CNodeWithArgs::k_dependent() const
{
    bool ret = std::any_of(arg.begin(), arg.end(), [](RNode a) { return a && a->k_dependent(); });
    if (debug)
        std::cerr << ("CNodeWithArgs{" + tree_info() + "}.k_dep -> " + S(ret) + "\n");
    return ret;
}

bool CNodeWithArgs::j_dependent() const
{
    bool ret = std::any_of(arg.begin(), arg.end(), [](RNode a) { return a && a->j_dependent(); });
    if (debug)
        std::cerr << ("CNodeWithArgs{" + tree_info() + "}.j_dep -> " + S(ret) + "\n");
    return ret;
}

bool CNodeWithArgs::i_dependent() const
{
    bool ret = std::any_of(arg.begin(), arg.end(), [](RNode a) { return a && a->i_dependent(); });
    if (debug)
        std::cerr << ("CNodeWithArgs{" + tree_info() + "}.i_dep -> " + S(ret) + "\n");
    return ret;
}


//**************************************************************************************************
//* CNodeFun: function/operator node
//**************************************************************************************************

CNodeFun1::CNodeFun1(const CFunc* _fu, RNode _a0) : CNodeFun(1, _fu)
{
    arg[0] = _a0;
    if (debug)
        std::cerr << ("CNodeFun1 := " + tree_info() + "\n");
}

CNodeFun2::CNodeFun2(const CFunc* _fu, RNode _a0, RNode _a1) : CNodeFun(2, _fu)
{
    arg[0] = _a0;
    arg[1] = _a1;
    if (debug)
        std::cerr << ("CNodeFun2 := " + tree_info() + "\n");
}

CNodeFun3::CNodeFun3(const CFunc* _fu, RNode _a0, RNode _a1, RNode _a2) : CNodeFun(3, _fu)
{
    arg[0] = _a0;
    arg[1] = _a1;
    arg[2] = _a2;
    if (debug)
        std::cerr << ("CNodeFun1 := " + tree_info() + "\n");
}

//! Returns the result of function evaluation for given arguments.

RObj CNodeFun::tree_val(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeFun{" + tree_info() + "}.tree_val ...\n");
    string basetypes;
    const CTypedFunc* tf = nullptr;
    try {
        // evaluate arguments, determine if all results are scalar.
        vector<RObj> pa(narg);
        basetypes = "";
        bool is_scalar = true;
        for (int iarg = 0; iarg < narg; ++iarg) {
            pa[iarg] = arg[iarg]->tree_val(ctx);
            basetypes += pa[iarg]->base_type();
            if (pa[iarg]->is_vec())
                is_scalar = false;
        }

        tf = fu->find_tyfu(basetypes, ctx.want_error);
        if (!tf)
            throw "function " + fu->tag + " not implemented for argument types " + basetypes;
        funcPtr f = tf->f;

        if (is_scalar) {
            if (tf->outtype == "0") {
                if (tf->intypes == "i")
                    (*(func_0_i)(f))(pa[0]->to_i());
                else if (tf->intypes == "d")
                    (*(func_0_d)(f))(pa[0]->to_r());
                else if (tf->intypes == "s")
                    (*(func_0_s)(f))(pa[0]->to_s());
                else
                    throw S("BUG: unexpected intypes");
                return RObjStr(new CObjStr("__nullptr__"));
            } else if (tf->outtype == "i") {
                int val;
                if (tf->intypes == "i")
                    val = (*(func_i_i)(f))(pa[0]->to_i());
                else if (tf->intypes == "d")
                    val = (*(func_i_d)(f))(pa[0]->to_r());
                else if (tf->intypes == "s")
                    val = (*(func_i_s)(f))(pa[0]->to_s());
                else if (tf->intypes == "ii")
                    val = (*(func_i_ii)(f))(pa[0]->to_i(), pa[1]->to_i());
                else if (tf->intypes == "is")
                    val = (*(func_i_is)(f))(pa[0]->to_i(), pa[1]->to_s());
                else if (tf->intypes == "dd")
                    val = (*(func_i_dd)(f))(pa[0]->to_r(), pa[1]->to_r());
                else if (tf->intypes == "si")
                    val = (*(func_i_si)(f))(pa[0]->to_s(), pa[1]->to_i());
                else if (tf->intypes == "iii")
                    val = (*(func_i_iii)(f))(pa[0]->to_i(), pa[1]->to_i(), pa[2]->to_i());
                else
                    throw S("BUG: unexpected intypes");
                return RObjInt(new CObjInt(val));

            } else if (tf->outtype == "d") {
                double val;
                if (tf->intypes == "i")
                    val = (*(func_d_i)(f))(pa[0]->to_i());
                else if (tf->intypes == "d")
                    val = (*(func_d_d)(f))(pa[0]->to_r());
                else if (tf->intypes == "s")
                    val = (*(func_d_s)(f))(pa[0]->to_s());
                else if (tf->intypes == "ii")
                    val = (*(func_d_ii)(f))(pa[0]->to_i(), pa[1]->to_i());
                else if (tf->intypes == "dd")
                    val = (*(func_d_dd)(f))(pa[0]->to_r(), pa[1]->to_r());
                else if (tf->intypes == "sd")
                    val = (*(func_d_sd)(f))(pa[0]->to_s(), pa[1]->to_r());
                else if (tf->intypes == "ddd")
                    val = (*(func_d_ddd)(f))(pa[0]->to_r(), pa[1]->to_r(), pa[2]->to_r());
                else
                    throw S("BUG: unexpected intypes");
                return RObjDbl(new CObjDbl(val));
            } else if (tf->outtype == "e") {
                double val, err;
                if (tf->intypes == "e")
                    (*(func_e_e)(f))(val, err, pa[0]->to_r(), pa[0]->to_dr());
                else if (tf->intypes == "dd")
                    (*(func_e_dd)(f))(val, err, pa[0]->to_r(), pa[1]->to_r());
                else if (tf->intypes == "ee")
                    (*(func_e_ee)(f))(
                        val, err, pa[0]->to_r(), pa[0]->to_dr(), pa[1]->to_r(), pa[1]->to_dr());
                else if (tf->intypes == "se")
                    (*(func_e_se)(f))(val, err, pa[0]->to_s(), pa[1]->to_r(), pa[1]->to_dr());
                else if (tf->intypes == "eee")
                    (*(func_e_eee)(f))(
                        val, err, pa[0]->to_r(), pa[0]->to_dr(), pa[1]->to_r(), pa[1]->to_dr(),
                        pa[2]->to_r(), pa[2]->to_dr());
                else
                    throw S("BUG: unexpected intypes");
                return RObjEnu(new CObjEnu(val, err));
            } else if (tf->outtype == "s") {
                string val;
                if (tf->intypes == "s")
                    val = (*(func_s_s)(f))(pa[0]->to_s());
                else if (tf->intypes == "si")
                    val = (*(func_s_si)(f))(pa[0]->to_s(), pa[1]->to_i());
                else if (tf->intypes == "ss")
                    val = (*(func_s_ss)(f))(pa[0]->to_s(), pa[1]->to_s());
                else
                    throw S("BUG: unexpected intypes");
                return PObjStr(new CObjStr(val));
            } else
                throw S("BUG: unexpected outtype");
        } else { // is_vec()
            // determine length of input vectors
            int n = 1;
            for (int iarg = 0; iarg < narg; ++iarg) {
                int na = pa[iarg]->size();
                if (n == 1)
                    n = na;
                else if (na != 1 && na != n)
                    throw S("vector arguments have different sizes");
            }
            if (tf->outtype == "i") {
                PObjVecInt ret(new CObjVecInt(n));
                if (tf->intypes == "i")
                    for (int i = 0; i < n; ++i)
                        ret->v[i] = (*(func_i_i)(f))(pa[0]->to_i(i));
                else if (tf->intypes == "d")
                    for (int i = 0; i < n; ++i)
                        ret->v[i] = (*(func_i_d)(f))(pa[0]->to_r(i));
                else if (tf->intypes == "ii")
                    for (int i = 0; i < n; ++i)
                        ret->v[i] = (*(func_i_ii)(f))(pa[0]->to_i(i), pa[1]->to_i(i));
                else if (tf->intypes == "dd")
                    for (int i = 0; i < n; ++i)
                        ret->v[i] = (*(func_i_dd)(f))(pa[0]->to_r(i), pa[1]->to_r(i));
                else if (tf->intypes == "iii")
                    for (int i = 0; i < n; ++i)
                        ret->v[i] =
                            (*(func_i_iii)(f))(pa[0]->to_i(i), pa[1]->to_i(i), pa[2]->to_i(i));
                else
                    throw S("BUG: unexpected intypes");
                return ret;
            } else if (tf->outtype == "d") {
                PObjVecDbl ret(new CObjVecDbl(n));
                if (tf->intypes == "i")
                    for (int i = 0; i < n; ++i)
                        ret->v[i] = (*(func_d_i)(f))(pa[0]->to_i(i));
                else if (tf->intypes == "d")
                    for (int i = 0; i < n; ++i)
                        ret->v[i] = (*(func_d_d)(f))(pa[0]->to_r(i));
                else if (tf->intypes == "ii")
                    for (int i = 0; i < n; ++i)
                        ret->v[i] = (*(func_d_ii)(f))(pa[0]->to_i(i), pa[1]->to_i(i));
                else if (tf->intypes == "dd")
                    for (int i = 0; i < n; ++i)
                        ret->v[i] = (*(func_d_dd)(f))(pa[0]->to_r(i), pa[1]->to_r(i));
                else if (tf->intypes == "ddd")
                    for (int i = 0; i < n; ++i)
                        ret->v[i] =
                            (*(func_d_ddd)(f))(pa[0]->to_r(i), pa[1]->to_r(i), pa[2]->to_r(i));
                else
                    throw S("BUG: unexpected intypes");
                return ret;
            } else if (tf->outtype == "e") {
                PObjVecEnu ret(new CObjVecEnu(n));
                if (tf->intypes == "e")
                    for (int i = 0; i < n; ++i)
                        (*(func_e_e)(f))(ret->v[i], ret->dv[i], pa[0]->to_r(i), pa[0]->to_dr(i));
                else if (tf->intypes == "dd")
                    for (int i = 0; i < n; ++i)
                        (*(func_e_dd)(f))(ret->v[i], ret->dv[i], pa[0]->to_r(i), pa[1]->to_r(i));
                else if (tf->intypes == "ee")
                    for (int i = 0; i < n; ++i)
                        (*(func_e_ee)(f))(
                            ret->v[i], ret->dv[i], pa[0]->to_r(i), pa[0]->to_dr(i), pa[1]->to_r(i),
                            pa[1]->to_dr(i));
                else if (tf->intypes == "eee")
                    for (int i = 0; i < n; ++i)
                        (*(func_e_eee)(f))(
                            ret->v[i], ret->dv[i], pa[0]->to_r(i), pa[0]->to_dr(i), pa[1]->to_r(i),
                            pa[1]->to_dr(i), pa[2]->to_r(i), pa[2]->to_dr(i));
                else
                    throw S("BUG: unexpected intypes");
                return ret;
            } else
                throw S("BUG: unexpected outtype");
        }
        throw S("BUG: unexpected case");
    } catch (string& ex) {
        throw ex + "\n called by function node " + tree_info() + " {intypes='" + basetypes + "', "
            + (tf ? ("cast to '" + tf->intypes + "', outtype='" + tf->outtype + "', ") : "")
            + "context: " + ctx.context_info() + "}";
    }
}


CCoord CNodeFun1::node_coord(int k) const
{
    return fu->coord(arg[0]->node_coord(k));
}

CCoord CNodeFun2::node_coord(int k) const
{
    return fu->coord(arg[0]->node_coord(k), arg[1]->node_coord(k));
}

CCoord CNodeFun3::node_coord(int k) const
{
    return fu->coord(arg[0]->node_coord(k), arg[1]->node_coord(k), arg[2]->node_coord(k));
}

//! Returns string representation of this node.

string CNodeFun1::tree_info() const
{
    return fu->tag + "(" + arg[0]->tree_info() + ")";
}

//! Returns string representation of this node.

string CNodeFun2::tree_info() const
{
    return fu->tag + "(" + arg[0]->tree_info() + "," + arg[1]->tree_info() + ")";
}

//! Returns string representation of this node.

string CNodeFun3::tree_info() const
{
    return fu->tag + "(" + arg[0]->tree_info() + "," + arg[1]->tree_info() + ","
        + arg[2]->tree_info() + ")";
}


//**************************************************************************************************
//* CNodeRange: range node
//**************************************************************************************************

//! Evaluates the range in context, and returns a vector object pointer.

RObj CNodeRange::tree_val(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeRange{" + tree_info() + "}.tree_val ...\n");
    // evaluate arguments, determine types, assert they are scalar.
    vector<RObj> pa(narg);
    bool all_int = true;
    for (int iarg = 0; iarg < narg; ++iarg) {
        if (arg[iarg])
            pa[iarg] = arg[iarg]->tree_val(ctx);
        else if (iarg == 0)
            pa[iarg] = RObj(new CObjInt(0)); // default begin
        else if (iarg == 1)
            pa[iarg] = RObj(new CObjInt(ctx.nv - 1)); // default end
        else if (iarg == 2)
            pa[iarg] = RObj(new CObjInt(1)); // default step
        if (pa[iarg]->base_type() != 'i')
            all_int = false;
        if (pa[iarg]->is_vec())
            throw "vectorial argument in range " + tree_info();
    }

    if (all_int) {
        PObjVecInt ret(new CObjVecInt);
        for (int i = pa[0]->to_i(); (inclusive_end ? i <= pa[1]->to_i() : i < pa[1]->to_i());
             i += pa[2]->to_i())
            ret->v.push_back(i);
        return ret;
    } else {
        PObjVecDbl ret(new CObjVecDbl);
        for (double r = pa[0]->to_r(); (inclusive_end ? r <= pa[1]->to_r() : r < pa[1]->to_r());
             r += pa[2]->to_r())
            ret->v.push_back(r);
        return ret;
    }
}

CCoord CNodeRange::node_coord(int k) const
{
    return CCoord("range", "");
}

//! Returns string representation of this node.

string CNodeRange::tree_info() const
{
    return "range(" + (arg[0] ? arg[0]->tree_info() : "0") + ","
        + (arg[1] ? arg[1]->tree_info() : "end") + "," + (arg[2] ? arg[2]->tree_info() : "1") + ")";
}


//**************************************************************************************************
//* CNodeList
//**************************************************************************************************

CNodeList::CNodeList(RNode _incremand, RNode _increment) : CNodeWithArgs(1)
{
    if (auto old = PCAST<const CNodeList>(_incremand)) {
        arg = old->args();
    } else {
        arg[0] = _incremand;
    }
    if (_increment)
        arg.push_back(_increment);
    narg = arg.size();
}


//! Evaluates the range in context, and returns a vector object pointer.

RObj CNodeList::tree_val(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeList{" + tree_info() + "}.tree_val ...\n");
    // evaluate arguments, determine types
    vector<RObj> pa(narg);
    bool all_int = true;
    bool all_dbl = true;
    bool has_enu = false;
    for (int iarg = 0; iarg < narg; ++iarg) {
        if (arg[iarg])
            pa[iarg] = arg[iarg]->tree_val(ctx);
        char t = pa[iarg]->base_type();
        if (t != 'i')
            all_int = false;
        if (t != 'i' && t != 'd' && t != 'e')
            all_dbl = false;
        if (t == 'e')
            has_enu = true;
    }

    if (all_int) {
        PObjVecInt ret(new CObjVecInt);
        for (int iarg = 0; iarg < narg; ++iarg) {
            if (RObjVecInt rav = PCAST<const CObjVecInt>(pa[iarg]))
                for (int i = 0; i < rav->size(); ++i)
                    ret->v.push_back(rav->to_i(i));
            else
                ret->v.push_back(pa[iarg]->to_i());
        }
        return ret;
    } else if (all_dbl) {
        if (has_enu) {
            PObjVecEnu ret(new CObjVecEnu);
            for (int iarg = 0; iarg < narg; ++iarg) {
                if (RObjVecNum rav = PCAST<const CObjVecNum>(pa[iarg])) {
                    for (int i = 0; i < rav->size(); ++i) {
                        ret->v.push_back(rav->to_r(i));
                        ret->dv.push_back(rav->to_dr(i));
                    }
                } else {
                    ret->v.push_back(pa[iarg]->to_r());
                    ret->dv.push_back(pa[iarg]->to_dr());
                }
            }
            return ret;
        } else {
            PObjVecDbl ret(new CObjVecDbl);
            for (int iarg = 0; iarg < narg; ++iarg) {
                if (RObjVecNum rav = PCAST<const CObjVecNum>(pa[iarg]))
                    for (int i = 0; i < rav->size(); ++i)
                        ret->v.push_back(rav->to_r(i));
                else
                    ret->v.push_back(pa[iarg]->to_r());
            }
            return ret;
        }
    } else {
        throw S("list-of-any not yet implemented");
    }
}

CCoord CNodeList::node_coord(int k) const { return CCoord("list", ""); }

//! Returns string representation of this node.

string CNodeList::tree_info() const
{
    string ret = "list(";
    for (int iarg = 0; iarg < narg; ++iarg) {
        ret += arg[iarg]->tree_info();
        if (iarg < narg - 1)
            ret += ",";
    }
    return ret;
}


//**************************************************************************************************
//* CNodeGeni: generalized integral node
//**************************************************************************************************


CNodeGeni::CNodeGeni(const class CGeni* _geni, RRef _ref) : CNodeWithArgs(_geni->narg), geni(_geni)
{
    arg.resize(narg);
    for (int iarg = 0; iarg < narg; ++iarg) {
        RRef ref = _ref ? _ref : RRef(new CRef3());
        if (geni->default_args[iarg] == "x")
            arg[iarg] = RNode(new CNodePointX(ref));
        else if (geni->default_args[iarg] == "y")
            arg[iarg] = RNode(new CNodePointY(ref));
        else if (geni->default_args[iarg] == "expr")
            throw S("syntax error: functional(expr) does not have default argument");
        else
            throw S("BUG: unexpected default arg");
    }
}

CNodeGeni::CNodeGeni(const class CGeni* _geni, RNode _a0) : CNodeWithArgs(1), geni(_geni)
{
    if (geni->narg != 1)
        throw "syntax error: generalized integral '" + geni->name + "' needs " + S(geni->narg)
            + " arguments, not 1";
    arg[0] = _a0;
}

CNodeGeni::CNodeGeni(const class CGeni* _geni, RNode _a0, RNode _a1) : CNodeWithArgs(2), geni(_geni)
{
    if (geni->narg != 2)
        throw "syntax error: generalized integral '" + geni->name + "' needs " + S(geni->narg)
            + " arguments, not 2";
    arg[0] = _a0;
    arg[1] = _a1;
}

//! A functional of 1,2,.. arguments (including operators like +,-,.. ).

RObj CNodeGeni::tree_val(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeGeni{" + tree_info() + "}.tree_val ...\n");
    try {
        if (ctx.dim == CContext::_VT)
            throw "Unexpected evaluation in curve context {context: " + ctx.context_info() + "}";
        CContext myctx(ctx);
        myctx.dim = CContext::_VI;
        myctx.nv = -1;
        // Evaluate arguments, and determine size of is_vec() arguments.
        vector<RObj> pa(narg);
        vector<RObjVecNum> pav(narg);
        int n = 0;
        for (int iarg = 0; iarg < narg; ++iarg) {
            pa[iarg] = arg[iarg]->tree_val(myctx);
            pav[iarg] = PCAST<const CObjVecNum>(pa[iarg]);
            if (!pav[iarg])
                throw "argument " + S(iarg) + " of generalized integral " + tree_info()
                    + " is scalar: " + pa[iarg]->result_info();
            if (n == 0)
                n = pav[iarg]->size();
            else if (pav[iarg]->size() != n)
                throw S("vector arguments have different size");
        }
        // Now evaluate the function.
        RObjNum val = geni->eval(n, pav); // TODO: compute errors
        if (ctx.dim == CContext::_1)
            return val;
        else if (ctx.dim == CContext::_VI)
            return val->to_vec(ctx.nv);
        else
            throw "BUG: unforeseen dimension {context: " + ctx.context_info() + "}";

    } catch (string& ex) {
        throw ex + "\n called by integral node " + tree_info() + " {context: " + ctx.context_info()
            + "}";
    }
}

CCoord CNodeGeni::node_coord(int k) const
{
    string rnam = geni->name + "(";
    vector<string> argunit;
    for (int iarg = 0; iarg < narg; ++iarg) {
        CCoord carg = arg[iarg]->node_coord(k);
        argunit.push_back(carg.unit);
        rnam += carg.name;
        rnam += iarg == narg - 1 ? ")" : ",";
    }
    string runit;
    if (geni->unit_hint == "?") {
        runit = "";
        cout << "TODO: hard-code unit for this generalized integral\n";
    } else if (geni->unit_hint == "-") {
        runit = "";
    } else if (geni->unit_hint == "a") {
        if (narg < 1)
            throw "BUG: unit hint '" + geni->unit_hint + "'  inconsistent with narg=" + S(narg);
        runit = argunit[0];
    } else if (geni->unit_hint == "ab") {
        if (narg < 2)
            throw "BUG: unit hint '" + geni->unit_hint + "'  inconsistent with narg=" + S(narg);
        runit = argunit[0] + "*" + argunit[1];
    } else {
        throw "BUG: unforeseen unit hint '" + geni->unit_hint + "'";
    }
    return CCoord(rnam, runit);
}


//! Returns string representation of this node.

string CNodeGeni::tree_info() const
{
    string ret = geni->name + "(";
    for (int iarg = 0; iarg < narg; ++iarg)
        ret += arg[iarg]->tree_info() + (iarg < narg - 1 ? "," : ")");
    return ret;
}


//**************************************************************************************************
//* CNodeCvin: curve functional node
//**************************************************************************************************


CNodeCvin::CNodeCvin(const class CCvin* _cvin, RRef _ref, RNode _a0, RNode _a1)
    : CNodeWithArgs(2), cvin(_cvin), ref(_ref)
{
    if (cvin->narg != 2)
        throw "syntax error: curve functional '" + cvin->name + "' needs " + S(cvin->narg)
            + " arguments, not 2";
    arg[0] = _a0;
    arg[1] = _a1;
    if (i_dependent())
        throw S("Invalid i dependence");
}

RObj CNodeCvin::tree_val(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeCvin{" + tree_info() + "}.tree_val ...\n");
    try {
        if (ctx.dim != CContext::_1)
            throw "Unexpected is_vec() {context: " + ctx.context_info() + "}";
        CContext myctx(ctx);
        myctx.dim = CContext::_1;
        // Evaluate arguments.
        vector<double> a(narg);
        for (int iarg = 0; iarg < narg; ++iarg) {
            RObj pa = arg[iarg]->tree_val(myctx);
            try {
                a[iarg] = pa->to_r();
            } catch (string& ex) {
                throw "argument " + S(iarg) + " of curve functional " + tree_info()
                    + " is not a scalar number: " + pa->result_info() + " => " + ex;
            }
        }
        int k = ref->get_k(ctx);
        const COlo* f = SMem::instance()->mem_get(k);
        const COlc* fc = dynamic_cast<const COlc*>(f);
        if (!fc)
            throw S("This is not a curve file => cannot integrate");
        int j = PCAST<const CRef2>(ref)->get_j(ctx, f->nJ());
        // now evaluate the function
        double val;
        cvin->eval(&val, nullptr, fc, k, j, a); // TODO: compute errors?
        return PObjDbl(new CObjDbl(val));
    } catch (string& ex) {
        throw ex + " in NodeCvin " + tree_info() + " {context: " + ctx.context_info() + "}";
    }
}

CCoord CNodeCvin::node_coord(int k) const
{
    //*
    //*     CCoord r[maxarg];
    //*     for ( int iarg=0; iarg<narg; ++iarg )
    //*         arg[iarg]->set_coord( r[iarg], k );
    //*     if      ( const CFunc1* fu = dynamic_cast<const CFunc1*>( fun ) )
    //*         ret = fu->coord( r+0);
    //*     else if ( const CFunc2* fu = dynamic_cast<const CFunc2*>( fun ) )
    //*         ret = fu->coord( r+0, r+1 );
    //*     else if ( const CFunc3* fu = dynamic_cast<const CFunc3*>( fun ) )
    //*         ret = fu->coord( r+0, r+1, r+2 );
    return CCoord("TODO:CvinCoord", "");
}

//! Returns string representation of this node.

string CNodeCvin::tree_info() const
{
    string ret = cvin->name + "(..)";
    for (int iarg = 0; iarg < narg; ++iarg)
        ret += arg[iarg]->tree_info() + (iarg < narg - 1 ? "," : ")");
    return ret;
}


//**************************************************************************************************
//* CNodeVal: end node containing a numeric value
//**************************************************************************************************

CNodeVal::CNodeVal(double _val) : val(PObjDbl(new CObjDbl(_val))) {}

CNodeVal::CNodeVal(int _val) : val(PObjInt(new CObjInt(_val))) {}

bool CNodeVal::looks_like_indices() const { return PCAST<const CObjInt>(val) != nullptr; }

RObj CNodeVal::tree_val(const CContext& ctx) const {
    if (debug)
        std::cerr << ("CNodeVal{" + tree_info() + "}.tree_val -> " + val->to_s() + "\n");
    return val;
}

CCoord CNodeVal::node_coord(int k) const { return CCoord(val->to_s(), ""); }

string CNodeVal::tree_info() const { return "val(" + val->to_s() + ")"; }


//**************************************************************************************************
//* CNodeIdx: index nodes
//**************************************************************************************************


RObj CNodeIdxI::tree_val(const CContext& ctx) const
{
    if (ctx.dim == CContext::_1) {
        if (ctx.i == -1)
            throw S("missing index i");
        RObjInt ret(new CObjInt(ctx.i));
        if (debug)
            std::cerr << ("CNodeIdxI{" + tree_info() + "}.tree_val[dim=1] -> "
                          + ret->to_s() + "\n");
        return ret;
    } else if (ctx.dim == CContext::_VI) {
        PObjVecInt pret(new CObjVecInt(ctx.nv));
        for (int i = 0; i < ctx.nv; ++i)
            pret->v[i] = i;
        if (debug)
            std::cerr << ("CNodeIdxI{" + tree_info() + "}.tree_val[dim=VI] -> "
                          + pret->to_s() + "\n");
        return pret;
    } else
        throw S("index i not defined in this context");
}


RObj CNodeIdxJ::tree_val(const CContext& ctx) const
{
    RObj ret = RObjInt(new CObjInt(ctx.j));
    if (debug)
        std::cerr << ("CNodeIdxJ{" + tree_info() + "}.tree_val -> " + ret->to_s() + "\n");
    return ret;
}


RObj CNodeIdxK::tree_val(const CContext& ctx) const
{
    RObj ret = RObjInt(new CObjInt(ctx.k));
    if (debug)
        std::cerr << ("CNodeIdxK{" + tree_info() + "}.tree_val -> " + ret->to_s() + "\n");
    return ret;
}


//**************************************************************************************************
//* CNodeSession: session-wide workspace information
//**************************************************************************************************

RObj CNodeSessionNK::tree_val(const CContext& ctx) const
{
    RObj ret = RObjInt(new CObjInt(SMem::instance()->mem_size()));
    if (debug)
        std::cerr << ("CNodeNK{" + tree_info() + "}.tree_val -> " + ret->to_s() + "\n");
    return ret;
}


//**************************************************************************************************
//* CNodeFile: file-wide information
//**************************************************************************************************

RObj CNodeFile::tree_val(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeFile{" + tree_info() + "}.tree_val ...\n");
    if (ctx.dim == CContext::_VI) {
        CContext myctx = ctx;
        myctx.dim = CContext::_1;
        PObjVecObj pret(new CObjVecObj());
        for (int ii = 0; ii < ctx.nv; ++ii) {
            myctx.i = ii;
            pret->v.push_back(tree_val_scalar(myctx));
        }
        return pret->to_vecnum();
    } else {
        return tree_val_scalar(ctx);
    }
}

CCoord CNodeFile::node_coord(int _k) const
{
    int k;
    if (ref->tk) {
        if (ref->tk->i_dependent() || ref->tk->j_dependent())
            return CCoord(name() + ref->ref_info(), "");
        CContext ctx(_k);
        k = ref->tk->tree_val_idx(ctx, "k");
    } else {
        k = _k;
    }
    const COlo* f = SMem::instance()->mem_get(k);
    return file_coord(f);
}

RObj CNodeFileNJ::tree_val_scalar(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeFileNJ{" + tree_info() + "}.tree_val_scalar ...\n");
    const COlo* f = SMem::instance()->mem_get(ref->get_k(ctx));
    return RObjInt(new CObjInt(f->nJ()));
}

RObj CNodeFileR::tree_val_scalar(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeFileR{" + tree_info() + "}.tree_val_scalar ...\n");
    const COlo* f = SMem::instance()->mem_get(ref->get_k(ctx));
    if (num < 0 || num >= f->RPar.size())
        throw "file has no parameter r" + S(num);
    return RObjDbl(new CObjDbl(f->RPar[num].val));
}

CCoord CNodeFileR::file_coord(const COlo* f) const
{
    if (num < 0 || num >= f->RPar.size())
        throw "file has no parameter r" + S(num);
    return f->RPar[num].Co;
}


//**************************************************************************************************
//* CNodeSlice/Spec/Curve: slice-wide information
//**************************************************************************************************


RObj CNodeSliceZ::tree_val_scalar(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeSliceZ{" + tree_info() + "}.tree_val_scalar ...\n");
    const COlo* f = SMem::instance()->mem_get(ref->get_k(ctx));
    int j = PCAST<const CRef2>(ref)->get_j(ctx, f->nJ());
    const CSlice* s = f->V[j].get();
    if (num < 0 || num >= s->z.size())
        throw "z" + S(num) + " requested, while only " + S(s->z.size()) + " z entries are defined";
    return s->z[num];
}

CCoord CNodeSliceZ::file_coord(const COlo* f) const
{
    if (num < 0 || num >= f->ZCo.size())
        throw "z" + S(num) + " requested, while only " + S(f->ZCo.size())
            + " z coordinates are defined";
    return f->ZCo[num];
}

RObj CNodeSpecNI::tree_val_scalar(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeSpecNI{" + tree_info() + "}.tree_val_scalar ...\n");
    const COlo* f = SMem::instance()->mem_get(ref->get_k(ctx));
    const COld* fd = dynamic_cast<const COld*>(f);
    if (!fd)
        throw f->type() + " file has no property " + tree_info();
    int j = PCAST<const CRef2>(ref)->get_j(ctx, f->nJ());
    return RObjInt(new CObjInt(fd->VS(j)->size()));
}

RObj CNodeCurve::tree_val_scalar(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeCurve{" + tree_info() + "}.tree_val_scalar ...\n");
    const COlo* f = SMem::instance()->mem_get(ref->get_k(ctx));
    const COlc* fc = dynamic_cast<const COlc*>(f);
    if (!fc)
        throw f->type() + " file has no property " + tree_info();
    int j = PCAST<const CRef2>(ref)->get_j(ctx, f->nJ());
    return curve_val_scalar(fc->VC(j));
}


RObj CNodeCurveP::curve_val_scalar(const CCurve* c) const
{
    if (debug)
        std::cerr << ("CNodeCurveP{" + tree_info() + "}.tree_val_scalar ...\n");
    if (num < 0 || num >= c->P.size())
        throw "curve has no p" + S(num);
    if (double err = c->dP(num))
        return RObjEnu(new CObjEnu(c->P[num], err));
    return RObjDbl(new CObjDbl(c->P[num]));
}

CCoord CNodeCurveP::file_coord(const COlo* f) const
{
    const COlc* fc = dynamic_cast<const COlc*>(f);
    if (!fc)
        throw f->type() + " file has no property " + tree_info();
    if (num < 0 || num >= fc->PCo.size())
        throw "p" + S(num) + " requested, while only " + S(fc->PCo.size())
            + " z coordinates are defined";
    return fc->PCo[num];
}


RObj CNodeCurveOutcome::curve_val_scalar(const CCurve* c) const
{
    return RObjInt(new CObjInt(c->fitOutcome));
}

RObj CNodeCurveNfev::curve_val_scalar(const CCurve* c) const
{
    return RObjInt(new CObjInt(c->nFunctionEvaluations));
}

RObj CNodeCurveChi2::curve_val_scalar(const CCurve* c) const
{
    return RObjDbl(new CObjDbl(c->fitChi2));
}

RObj CNodeCurveR2::curve_val_scalar(const CCurve* c) const
{
    return RObjDbl(new CObjDbl(c->fitR2));
}


RObj CNodePoint::tree_val(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodePoint{" + tree_info() + "}.tree_val ...\n");
    auto ref3 = PCAST<const CRef3>(ref);
    if (!ref3)
        throw S("BUG: NodePoint has no Ref3");
    if (ctx.dim == CContext::_VI) {
        if (ref->k_depends_on_i() || ref3->j_depends_on_i()
            || (ref3->ti && ref3->ti->i_dependent())) {
            if (ctx.nv < 0)
                throw S("BUG: i-dependent K or J with unspecified return array length");
            if (!ref3->ti) // Could be easily supported, but is forbidden to prevent accidental use
                throw S("Explicit I required with i-dependent K or J");
            PObjVecObj tmp(new CObjVecObj());
            CContext myctx(ctx);
            myctx.dim = CContext::_1;
            for (int i = 0; i < ctx.nv; ++i) {
                myctx.i = i;
                tmp->v.push_back(tree_val_point(myctx, ref3->ti->tree_val_idx(myctx, "i")));
            }
            RObj ret = tmp->to_vecnum();
            if (debug)
                std::cerr << ("... CNodePoint{" + tree_info() + "}.tree_val[dim=VI,i2j||i2k] -> "
                              + ret->to_s() + "\n");
            return ret;
        } else if (!ref3->ti) {
            RObj ret = tree_val_vector(ctx);
            if (debug)
                std::cerr << ("... CNodePoint{" + tree_info() + "}.tree_val[dim=VI,i2nil] -> "
                              + ret->to_s() + "\n");
            return ret;
        } else {
            RObj ret = tree_val_point(ctx, ref3->ti->tree_val_idx(ctx, "i"));
            if (debug)
                std::cerr << ("... CNodePoint{" + tree_info() + "}.tree_val[dim=VI,i2i] -> "
                              + ret->to_s() + "\n");
            return ret;
        }
    } else {
        if (ref3->ti) {
            RObj ret = tree_val_point(ctx, ref3->ti->tree_val_idx(ctx, "i"));
            if (debug)
                std::cerr << ("... CNodePoint{" + tree_info() + "}.tree_val[dim!=VI,ti] -> "
                              + ret->to_s() + "\n");
            return ret;
        } else {
            RObj ret = tree_val_point(ctx, ctx.i);
            if (debug)
                std::cerr << ("... CNodePoint{" + tree_info() + "}.tree_val[dim!=VI,!ti] -> "
                              + ret->to_s() + "\n");
            return ret;
        }
    }
}


RObj CNodePoint::tree_val_point(const CContext& ctx, int i) const
{
    if (debug)
        std::cerr << ("CNodePoint{" + tree_info() + "}.tree_val_point -> ...\n");
    const COlo* f = SMem::instance()->mem_get(ref->get_k(ctx));
    const COld* fd = dynamic_cast<const COld*>(f);
    if (!fd)
        throw f->type() + " file has no property " + tree_info();
    int j = PCAST<const CRef2>(ref)->get_j(ctx, f->nJ());
    const CSpec* s = fd->VS(j);
    if (i < 0 || i >= s->size())
        throw "invalid data point index " + S(i) + ", spectrum has length " + S(s->size());
    RObj ret = spec_val_point(s, i, ctx.want_error);
    if (debug)
        std::cerr << ("... CNodePoint{" + tree_info() + "}.tree_val_point -> "
                      + ret->to_s() + "\n");
    return ret;
}


RObj CNodePoint::tree_val_vector(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodePoint{" + tree_info() + "}.tree_val_vector ...\n");
    if (ctx.dim != CContext::_VI)
        throw S("BUG: unexpected context in tree_val_vector");
    const COlo* f = SMem::instance()->mem_get(ref->get_k(ctx));
    const COld* fd = dynamic_cast<const COld*>(f);
    if (!fd)
        throw f->type() + " file has no property " + tree_info();
    int j = PCAST<const CRef2>(ref)->get_j(ctx, f->nJ());
    const CSpec* s = fd->VS(j);
    int n = ctx.nv;
    if (n < 0) // ctx.nv=-1 is used in integral operations
        n = s->size();
    else if (n > s->size())
        throw "requested " + S(n) + " points while spectrum " + S(j) + " only has " + S(s->size())
            + " points";
    RObj ret = spec_val_vector(s, n, ctx.want_error);
    if (debug)
        std::cerr << ("... CNodePoint{" + tree_info() + "}.tree_val_vector -> "
                      + ret->to_s() + "\n");
    return ret;
}


RObj CNodePointX::spec_val_point(const CSpec* s, int i, bool want_error) const
{
    RObj ret = RObjDbl(new CObjDbl(s->x[i]));
    if (debug)
        std::cerr << ("CNodePointX{" + tree_info() + "}.spec_val_point -> " + ret->to_s() + "\n");
    return ret;
}

RObj CNodePointX::spec_val_vector(const CSpec* s, int n, bool want_error) const
{
    RObj ret = RObjVecDbl(new CObjVecDbl(vector<double>(s->x.begin(), s->x.begin() + n)));
    if (debug)
        std::cerr << ("CNodePointX{" + tree_info() + "}.spec_val_vector -> " + ret->to_s() + "\n");
    return ret;
}

CCoord CNodePointX::file_coord(const COlo* f) const
{
    const COld* fd = dynamic_cast<const COld*>(f);
    if (!fd)
        throw f->type() + " file has no property " + tree_info();
    return fd->xco;
}


RObj CNodePointY::spec_val_point(const CSpec* s, int i, bool want_error) const
{
    RObj ret = want_error && s->has_dy()
        ? RObjEnu(new CObjEnu(s->y[i], s->dy[i]))
        : RObjDbl(new CObjDbl(s->y[i]));
    if (debug)
        std::cerr << ("CNodePointY{" + tree_info() + "}.spec_val_point -> " + ret->to_s() + "\n");
    return ret;
}

RObj CNodePointY::spec_val_vector(const CSpec* s, int n, bool want_error) const
{
    RObj ret = want_error && s->has_dy()
        ? RObjVecEnu(new CObjVecEnu(
              vector<double>(s->y.begin(), s->y.begin() + n),
              vector<double>(s->dy.begin(), s->dy.begin() + n)))
        : RObjVecDbl(new CObjVecDbl(vector<double>(s->y.begin(), s->y.begin() + n)));
    if (debug)
        std::cerr << ("CNodePointY{" + tree_info() + "}.spec_val_vector -> " + ret->to_s() + "\n");
    return ret;
}

CCoord CNodePointY::file_coord(const COlo* f) const
{
    const COld* fd = dynamic_cast<const COld*>(f);
    if (!fd)
        throw f->type() + " file has no property " + tree_info();
    return fd->yco;
}


RObj CNodePointDY::spec_val_point(const CSpec* s, int i, bool want_error) const
{
    RObj ret = RObjDbl(new CObjDbl(s->dy[i]));
    if (debug)
        std::cerr << ("CNodePointDY{" + tree_info() + "}.spec_val_point -> " + ret->to_s() + "\n");
    return ret;
}

RObj CNodePointDY::spec_val_vector(const CSpec* s, int n, bool want_error) const
{
    if (!s->has_dy())
        throw S("requested dy does not exist");
    RObj ret = RObjVecDbl(new CObjVecDbl(vector<double>(s->dy.begin(), s->dy.begin() + n)));
    if (debug)
        std::cerr << ("CNodePointDY{" + tree_info() + "}.spec_val_vector -> " + ret->to_s() + "\n");
    return ret;
}

CCoord CNodePointDY::file_coord(const COlo* f) const
{
    const COld* fd = dynamic_cast<const COld*>(f);
    if (!fd)
        throw f->type() + " file has no property " + tree_info();
    return CCoord("d" + fd->yco.name, fd->yco.unit);
}


//**************************************************************************************************
//* CNodeIdf: identifier node
//**************************************************************************************************

CNodeIdf::CNodeIdf(string _idf) : idf(_idf) {}

RObj CNodeIdf::tree_val(const CContext& ctx) const
{
    RObj ret = SVariRegistry::instance()->find(idf);
    if (!ret)
        throw "uninitialized variable " + idf;
    if (debug)
        std::cerr << ("CNodeIdf{" + tree_info() + "}.tree_val -> " + ret->to_s() + "\n");
    return ret;
}

CCoord CNodeIdf::node_coord(int k) const
{
    return CCoord(idf, ""); // primitiver gehts nicht
}


//**************************************************************************************************
//* CNodeDummy: function argument (t) node
//**************************************************************************************************

RObj CNodeDummy::tree_val(const CContext& ctx) const
{
    if (!(ctx.vt))
        throw "BUG: tree_val with *vt=0";
    RObjVecDbl pret(new CObjVecDbl(*(ctx.vt)));
    return pret;
}

//**************************************************************************************************
//* CNodeCev: curve evaluation node
//**************************************************************************************************


//! A curve evaluation: fc[k,j](arg).

RObj CNodeCev::tree_val(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeCev{" + tree_info() + "}.tree_val ...\n");
    try {
        // get k, curve, j from context:
        int k = ref->get_k(ctx);
        const COlc* fc = SMem::instance()->mem_get_C(k);
        int j = PCAST<const CRef2>(ref)->get_j(ctx, fc->nJ());

        // evaluate function argument:
        RObj farg = arg->tree_val(ctx);

        // evaluate the curve:
        if (RObjVecDbl pv_farg = PCAST<const CObjVecDbl>(farg)) {
            return fc->eval_curve(pv_farg->v, k, j, ctx.want_error);
        } else if (RObjNum pn_farg = PCAST<const CObjNum>(farg)) {
            return fc->eval_curve_scalar(pn_farg->to_r(), k, j, ctx.want_error);
        } else
            throw S("unexpected argument type");
    } catch (string& ex) {
        throw ex + " in NodeCev " + tree_info() + " {context: " + ctx.context_info() + "}";
    }
}

CCoord CNodeCev::node_coord(int k) const { return CCoord("c", ""); } // primitiver gehts nicht


//**************************************************************************************************
//* CNodeMixin: Base class for mixing in data from a "convolution" file
//**************************************************************************************************

RObj CNodeMixin::tree_val(const CContext& ctx) const
{
    if (debug)
        std::cerr << ("CNodeMixin{" + tree_info() + "}.tree_val ...\n");
    // Shift:
    RObj res_shift = shift->tree_val(ctx);
    RObjNum par = PCAST<const CObjNum>(res_shift);
    if (!par)
        throw S("shift is not a scalar");
    double theshift = par->to_r();

    // Resolution is given by context 'ctx':
    int kconv, jconv;
    const CSpec* sv;
    NCurveFile::get_conv(&sv, &kconv, &jconv, ctx.k, ctx.j);

    PObjVecDbl pret(ctx.want_error ? (new CObjVecDbl) : (new CObjVecEnu));

    if (kconv == -1) {
        return copy_theory(ctx, theshift);
    } else {
        double conv_norm; // normalization constant (integral of resolution)
        double conv_step; // step in x/t, if equidistant
        string info = "resolution " + S(kconv) + ", spectrum " + S(jconv);
        conv_norm = sv->norm(info);
        conv_step = sv->has_step(ctx.vt_step) ? ctx.vt_step : 0;
        if (!conv_step && !allow_slow_conv)
            throw S("grid is not equidistant; slow convolution is not allowed");
        return convolve(ctx, theshift, sv, conv_norm, conv_step);
    }
}


//**************************************************************************************************
//* CNodeConv: convoluted function node
//**************************************************************************************************

RObj CNodeConv::copy_theory(const CContext& ctx, double theshift) const
{
    return theory->tree_val(ctx);
}

RObjVecDbl CNodeConv::convolve(
    const CContext& ctx, double theshift, const CSpec* sv, double conv_norm, double conv_step) const
{
    int n = ctx.vt->size();
    int nv = sv->size();
    CContext cv_ctx(ctx);
    vector<double> tt; // theory grid
    cv_ctx.vt = &tt;
    PObjVecDbl ret;
    PObjVecEnu dret;
    if (ctx.want_error) {
        dret = PObjVecEnu(new CObjVecEnu(n));
        ret = dret;
    } else {
        ret = PObjVecDbl(new CObjVecDbl(n));
        dret = nullptr;
    }
    // Convolute theory with resolution.
    if (conv_step) { // accelerated version with equidistant grids
        // precompute theory on unified grid
        int ntt = n + nv - 1;
        tt.resize(ntt);
        double tt0 = (*(ctx.vt))[0] - sv->x[0] - theshift - (nv - 1) * conv_step;
        for (int ii = 0; ii < ntt; ++ii) // ii=(nv-1)+i-iv
            tt[ii] = tt0 + ii * conv_step;
        RObj res_theory = theory->tree_val(cv_ctx);
        // now convolve
        for (int i = 0; i < n; ++i) {
            ret->v[i] = 0;
            if (dret)
                dret->dv[i] = 0;
            for (int iv = 0; iv < nv; ++iv) {
                ret->v[i] += res_theory->to_r(nv - 1 + i - iv) * sv->y[iv];
                if (dret)
                    dret->dv[i] += fabs(res_theory->to_r(nv - 1 + i - iv)) * SQR(sv->dy[iv]);
            }
            ret->v[i] *= conv_step / conv_norm;
            if (dret)
                dret->dv[i] = sqrt(dret->dv[i]) * conv_step / conv_norm;
        }
    } else { // simplest version: non-equidistant x_out and x_res.
        ret->v.assign(n, 0.);
        if (dret)
            dret->dv.assign(n, 0.);
        tt.resize(n);
        for (int iv = 0; iv < nv; ++iv) {
            double dx; // variable step
            if (iv == 0)
                dx = sv->x[1] - sv->x[0];
            else if (iv == nv - 1)
                dx = sv->x[nv - 1] - sv->x[nv - 2];
            else
                dx = (sv->x[iv + 1] - sv->x[iv - 1]) / 2;
            for (int i = 0; i < n; ++i)
                tt[i] = (*(ctx.vt))[i] - sv->x[iv] - theshift;
            RObj res_theory = theory->tree_val(cv_ctx);
            for (int i = 0; i < n; ++i) {
                ret->v[i] += res_theory->to_r(i) * sv->y[iv] * dx;
                if (dret)
                    dret->dv[i] += res_theory->to_r(i) * SQR(sv->dy[iv]);
            }
        }
        for (int i = 0; i < n; ++i) {
            ret->v[i] /= conv_norm;
            if (dret)
                dret->dv[i] = sqrt(dret->dv[i]) / conv_norm;
        }
    }
    return ret;
}


//**************************************************************************************************
//* CNodePConv: convolution with primitive
//**************************************************************************************************

RObj CNodePConv::copy_theory(const CContext& ctx, double theshift) const
{
    throw "non-convolution not implemented for pconv";
    // to implement this efficiently, I would like to have CObj::operator+
    // for which we should go to shared pointers ...
}

RObjVecDbl CNodePConv::convolve(
    const CContext& ctx, double theshift, const CSpec* sv, double conv_norm, double conv_step) const
{
    if (conv_norm <= 0)
        throw "BUG: pconv called with conv_norm=" + S(conv_norm);
    int n = ctx.vt->size();
    int nv = sv->size();
    CContext cv_ctx(ctx);
    vector<double> tt;
    cv_ctx.vt = &tt;
    PObjVecDbl ret;
    PObjVecEnu dret;
    if (ctx.want_error) {
        dret = PObjVecEnu(new CObjVecEnu(n));
        ret = dret;
        // A check. We don't care about the cost O(nv), since this node is O(n*nv).
        for (int iv = 0; iv < nv; ++iv) {
            if (!std::isfinite(sv->dy[iv]))
                throw S("resolution function contains non-finite error");
            if (sv->dy[iv] < 0)
                throw S("resolution function contains negative error");
        }
    } else {
        ret = PObjVecDbl(new CObjVecDbl(n));
        dret = nullptr;
    }
    // Convolute theory with resolution.
    if (conv_step) { // accelerated version with equidistant grids
        // precompute primitive on unified grid
        int ntt = n + nv + 1;
        tt.resize(ntt);
        double tt0 = (*(ctx.vt))[0] - sv->x[0] - theshift - (nv - 0.5) * conv_step;
        for (int ii = 0; ii < ntt; ++ii) // ii=(nv-1)+i-iv
            tt[ii] = tt0 + ii * conv_step;
        RObj res_theory = theory->tree_val(cv_ctx);
        if (res_theory->size() != ntt)
            throw S("BUG: unexpected size conflict");
        for (int it = 0; it < ntt; ++it)
            if (!std::isfinite(res_theory->to_r(it)))
                throw S("theory returns infinite value");
        for (int it = 0; it < ntt - 1; ++it)
            if (res_theory->to_r(it) > res_theory->to_r(it + 1))
                throw "theory is decreasing from S(" + S(tt[it]) + "=" + S(res_theory->to_r(it))
                    + " to S(" + S(tt[it + 1]) + "=" + S(res_theory->to_r(it + 1));
        // now convolve resolution with differential of primitive
        for (int i = 0; i < n; ++i) {
            ret->v[i] = 0;
            if (dret)
                dret->dv[i] = 0;
            for (int iv = 0; iv < nv; ++iv) {
                double igral = res_theory->to_r(nv + i - iv) - res_theory->to_r(nv - 1 + i - iv);
                ret->v[i] += igral * sv->y[iv];
                if (dret)
                    dret->dv[i] += SQR(igral * sv->dy[iv]);
            }
            ret->v[i] /= conv_norm;
            if (!std::isfinite(ret->v[i]))
                throw S("BUG: pconv result has nonfinite result");
            if (dret) {
                if (!std::isfinite(dret->dv[i]))
                    throw S("BUG: pconv result has nonfinite variance");
                if (dret->dv[i] < 0)
                    throw S("BUG: pconv result has negative variance");
                dret->dv[i] = sqrt(dret->dv[i]) / conv_norm;
                if (!std::isfinite(dret->dv[i]))
                    throw S("BUG: pconv result has nonfinite error");
            }
        }
    } else { // simplest version: non-equidistant x_out and x_res.
        throw S("PCONV BROKEN");
        /*
                ret->v.assign( n, 0. );
                if( with_error )
                    ret->dv.assign( n, 0. );
                tt.resize(n);
                RObj res_theory_low;
                RObj res_theory_hig;
                double dx;
                for ( int iv=0; iv<nv; ++iv ) {
                    if      ( iv==0 )
                        dx = sv->x[1] - sv->x[0];
                    else if ( iv==nv-1 )
                        dx = sv->x[nv-1] - sv->x[nv-2];
                    else
                        dx = ( sv->x[iv+1] - sv->x[iv-1] ) / 2;
                    for ( int i=0; i<n; ++i ) {
                        tt[i] = (*(ctx.vt))[i] - sv->x[iv] - theshift - dx/2;
                        theory->tree_val( res_theory_low, cv_ctx );
                    }
                    for ( int i=0; i<n; ++i ) {
                        tt[i] = (*(ctx.vt))[i] - sv->x[iv] - theshift + dx/2;
                        theory->tree_val( res_theory_hig, cv_ctx );
                    }
                    for ( int i=0; i<n; ++i ) {
                        double igral = res_theory_hig->to_r(i)-res_theory_low->to_r(i);
                        ret->v[i] += igral * sv->y[iv] * dx;
                        if ( with_error )
                            ret->dv[i] += igral * SQR(sv->dy[iv]) * dx;
                    }
                }
                for ( int i=0; i<n; ++i ) {
                    ret->v[i] /= conv_norm;
                    if ( with_error )
                        ret->dv[i] = sqrt(ret->dv[i]) / conv_norm;
                }
        */
    }
    return ret;
}


//**************************************************************************************************
//* CNodeDirac: node that copies the resolution function
//**************************************************************************************************

RObj CNodeDirac::copy_theory(const CContext& ctx, double theshift) const
{
    // Naive representation of delta function,
    // for use when convolution is switched off.
    PObjVecDbl ret(new CObjVecDbl(ctx.vt->size()));
    for (int i = 0; i < ctx.vt->size(); ++i)
        ret->v[i] = fabs((*(ctx.vt))[i] - theshift) < 1e-2 ? 1e2 : 0;
    return ret;
}

//! Convolution or Dirac function: conv(f, shift), dirac(shift).

RObjVecDbl CNodeDirac::convolve(
    const CContext& ctx, double theshift, const CSpec* sv, double conv_norm, double conv_step) const
{
    int n = ctx.vt->size();
    PObjVecDbl ret;
    PObjVecEnu dret;
    if (ctx.want_error) {
        dret = PObjVecEnu(new CObjVecEnu(n));
        ret = dret;
    } else {
        ret = PObjVecDbl(new CObjVecDbl(n));
        dret = nullptr;
    }
    // Delta function copies resolution.
    vector<double> tt(n), yy(n), dd(n);
    // shift in t
    for (int i = 0; i < n; ++i)
        tt[i] = (*(ctx.vt))[i] - theshift;
    // interpolation
    try {
        sv->intpol(tt, &yy, dret ? &dd : nullptr);
    } catch (string& s) {
        throw "error while evaluating delta function with shift " + S(theshift) + ": " + s;
    }
    // normalization
    for (int i = 0; i < n; ++i) {
        ret->v[i] = yy[i] / conv_norm;
        if (dret)
            dret->dv[i] = dd[i] / conv_norm;
    }
    return ret;
}

CCoord CNodeDirac::node_coord(int k) const
{
    return CCoord("resol", "");
}
