//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  olf.hpp
//! \brief COlo, COld, COld online files

#ifndef OLF_H
#define OLF_H

#include "coord.hpp"
#include "ptr.hpp"

class CGenus;

//! Online object. Virtual base class for COld, COlc.

class COlo
{
public:
    // Meta data:
    string name;
    vector<string> lDoc;
    CCoord xco, yco;
    vector<CCoord> ZCo;
    vector<CParam> RPar;
    bool as_on_disk;

    // Core data:
    vector<PSlice> V;

    COlo();
    COlo& operator=(const COlo&) = delete;
    COlo(const COlo&) = delete;
    virtual ~COlo(); ///< magic: makes dynamic_cast possible

    void copy_meta_O_from_other(const COlo* from);
    void log_action(const string& line);
    void remove_z(int iz);
    string info_line(int j, const string& which) const;
    void check_integrity() const;

    virtual void copy_mainvec(const COlo* fin) = 0;

    int nJ() const { return V.size(); }
    int nZ() const { return ZCo.size(); }
    RObj z(int j, int iz) const;
    RObjVecObj zvec(int iz) const;
    virtual PSlice copy_slice(int j) const = 0;
    virtual POlo new_POlo() const = 0;
    virtual CCoord coord(const CGenus& genus) const = 0;
    virtual void set_coord(const CGenus& genus, CCoord& co) = 0;
    virtual string type() const = 0;
};


//! Online data file. This is Frida's data model.

class COld : public COlo
{
public:
    COld() : COlo() {}
    explicit COld(class COlc const* fc);

    // Functions that exist similarly for COld and COlc:
    void copy_meta_D_from_other(const COld* other);
    CSpec* VS(int j) const;
    PSlice copy_slice(int j) const;
    POlo new_POlo() const;
    POld new_POld() const;
    void copy_mainvec(const COlo* fin);

    // Overloaded functions:
    void set_coord(const CGenus& genus, CCoord& co);
    CCoord coord(const CGenus& genus) const;

    // Functions specific to data file:
    void purge_dy();

    int nPts(int j) const;
    int nPts() const; // 0 unless all spectra j have same nPts(j)
    bool dy_are_all_nonzero() const;
    bool has_dy() const;
    bool x_sorted() const;
    virtual string type() const { return "data"; }
};


//! Online curve file.

class COlc : public COlo
{
private:
    RObjVec eval_curve_expr(const vector<double>& vt, int k, int j, bool want_error) const;
    RObjVec eval_curve_syscall(const vector<double>& vt, int k, int j, bool want_error) const;

public:
    // Function:
    enum TEva { _EXPR, _SCR, _SCR_INTP } evaMode; ///< How curve is defined
    enum TScrInp { _STDIN, _CMDLIN } scrInpMode; ///< Where pars are read from
    string expr; ///< The curve's definition (input expression).
    RNode T; ///< The curve's definition (parsed tree).
    int kconv; ///< Index of resolution file to convolve with.
    int nP; ///< Number of parameters.

    // Refined description:
    vector<CCoord> PCo; ///< Names and units of fit parameters.

    // Fitting:
    int kd; ///< Index of data file to be fitted.
    enum TWgt {
        _LIN,
        _LOG,
        _VAR,
        _VARC,
        _VARD,
        _LAST_TWGT = _VARD
    } weighing; ///< Weight mode for fitting.
    static string wgtNames[];
    static TWgt name2wgt(string name);
    bool plot_to_grid; ///< Use grid from file 'kd' ?
    string prange_expr; ///< Restricts points to be plotted.
    RNode prange_T; ///< Parsed range_expr.
    double chi2; ///< Global chi^2

    COlc()
        : COlo()
        , evaMode(_EXPR)
        , scrInpMode(_STDIN)
        , kconv(-1)
        , kd(-1)
        , weighing(_VAR)
        , plot_to_grid(false)
        , prange_expr("")
        , chi2(0)
    {
    }

    // Functions that exist similarly for COld and COlc:
    void copy_meta_C_from_other(const COlc* other);
    CCurve* VC(int j) const;
    PSlice copy_slice(int j) const;
    POlo new_POlo() const;
    POlc new_POlc() const;
    void copy_mainvec(const COlo* fin);

    // Overloaded functions:
    void set_coord(const CGenus& genus, CCoord& co);
    CCoord coord(const CGenus& genus) const;

    // Pertinent functions:
    bool has_global_par() const;
    string weight_str() const;
    vector<string> info_settings() const;
    string info_table_header(const string& which) const;
    void print_info_table(const string& which) const;

    // The following nontrivial function are implemented in curve.cpp:
    void curve_set_defaults();
    void clear_fits_outcome();
    void parse_function(const string& expr);
    void curve_query(const string& quest);
    RObjVec eval_curve(const vector<double>& vt, int k, int j, bool want_error) const;
    RObj eval_curve_scalar(double arg, int k, int j, bool want_error) const;
    RObjVecInt eval_prange(const vector<double>& vt, int k, int j) const;
    virtual string type() const { return "curve"; }
};


//! Service functions for online objects.

namespace NOlo {
    int j2j(const COlo* fin, int jin, const COlo* fout );
} // namespace NOlo


#endif // OLF_H
