//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  plot.cpp
//! \brief NPlot: plot data and curves.

#include <cmath>
#include <cfloat>
#include <algorithm>
#include <boost/format.hpp>

#include "defs.hpp"

#include "../trivia/vector_ops.hpp"
#include "../trivia/file_ops.hpp" // TODO rm file creation from plot2D
#include "../plot/dualplot.hpp"

#include "config.hpp"
#include "curve.hpp"
#include "fsel.hpp"
#include "jsel.hpp"
#include "loop.hpp"
#include "expr.hpp"
#include "mem.hpp"
#include "obj.hpp"
#include "olf.hpp"
#include "plot.hpp"
#include "slice.hpp"
#include "variables.hpp"

using boost::format;

namespace {
    void determine_Xrange(CPlot* plot, vector<int>& JSel);
    void determine_Yrange(CPlot* plot, vector<int>& JSel);
    int plot_data(CPlot* plot, const COld* fd, int k, int j, int pstyle);
    int plot_curve_convolved(CPlot* plot, const COlc* fc, int k, int j, int cstyle);
    int plot_curve_to_grid(CPlot* plot, const COlc* fc, int k, int j, int cstyle);
    int plot_curve_equidist(CPlot* plot, const COlc* fc, int k, int j, int cstyle);
    int plot_curve_refine(CPlot* plot, const COlc* fc, int k, int j, int cstyle);
    void plot1D(class CPlot* plot, bool add, const string& mode);
    void plot2D(class CPlot* plot);

//**************************************************************************************************
//*  Subroutines: determine data ranges
//**************************************************************************************************

//! Loop over selected data files to determine x range.

    void determine_Xrange(CPlot* plot, vector<int>& JSel)
    {
        double inf, sup;
        FileIterator fiter(SFSel::instance()->selD());
        const vector<double>* vx;
        bool first = true;
        while (const COld* fd = fiter.nextD()) {
            // JSel->evaluate( 0, fd->nJ()-1 );
            for (int iv = 0; iv < JSel.size(); ++iv) {
                vx = &(fd->VS(JSel[iv])->x);
                for (int i = 0; i < vx->size(); ++i) {
                    if (plot->X.logflag && (*vx)[i] <= 0)
                        continue;
                    if (first) {
                        inf = (*vx)[i];
                        sup = (*vx)[i];
                        first = false;
                    } else {
                        if ((*vx)[i] < inf)
                            inf = (*vx)[i];
                        if ((*vx)[i] > sup)
                            sup = (*vx)[i];
                    }
                }
            }
        }
        if (first)
            throw S("x range is empty");
        plot->X.set_rounded_limits(inf, sup);
    }


//! Loop over selected files to determine y range for current x range.

    void determine_Yrange(CPlot* plot, vector<int>& JSel)
    {
        double inf = INFINITY, sup = -INFINITY;
        FileIterator fiter(SFSel::instance()->sel());
        while (const COlo* f = fiter.next()) {
            int k = fiter.k();
            const COlc* fc = dynamic_cast<const COlc*>(f);
            if (fc && fc->kconv != -1)
                continue;
            const COld* fd = dynamic_cast<const COld*>(f);
            // JSel->evaluate( 0, f->nJ()-1 );
            for (int iv = 0; iv < JSel.size(); ++iv) {
                int j = JSel[iv];
                CSpec* s;
                if (fd) {
                    s = fd->VS(j);
                } else {
                    PSpec tmp(new CSpec);
                    s = tmp.get();
                    plot->X.set_xgrid(s->x, 97);
                    RObjVec cu = fc->eval_curve(s->x, k, j, false);
                    s->y.resize(s->x.size());
                    for (int i = 0; i < s->size(); ++i)
                        s->y[i] = cu->to_r(i);
                }
                for (int i = 0; i < s->size(); i++) {
                    if (!plot->X.contains(s->x[i]))
                        continue;
                    if (!std::isfinite(s->y[i]) || plot->Y.logflag && s->y[i] <= 0)
                        continue;
                    if (s->y[i] < inf)
                        inf = s->y[i];
                    if (s->y[i] > sup)
                        sup = s->y[i];
                }
            }
        }
        if (inf == INFINITY || sup == -INFINITY)
            throw S("y range is empty");
        plot->Y.set_rounded_limits(inf, sup);
    }


//**************************************************************************************************
//*  Subroutines: plot one scan
//**************************************************************************************************

//! Plot slice j of data file fd; return number of points plotted.

    int plot_data(CPlot* plot, const COld* fd, int k, int j, int pstyle)
    {
        const CSpec* s = fd->VS(j);
        int n = s->x.size();
        if (n != s->y.size())
            throw S("BUG: plot: x.size<>y.size");
        // Filter: x,y -> xp,yp if inside frame
        int np = 0, nxl = 0, nxh = 0, nyl = 0, nyh = 0;
        vector<double> xp, yp, dyp;
        for (int i = 0; i < n; i++) {
            double x = s->x[i];
            if (!plot->X.contains(x)) {
                if (x <= plot->X.inf) {
                    x = plot->X.inf;
                    nxl++;
                }
                if (x >= plot->X.sup) {
                    x = plot->X.sup;
                    nxh++;
                }
                if (!plot->X.force)
                    continue;
            }
            double y = s->y[i];
            if (!plot->Y.contains(y)) {
                if (y <= plot->Y.inf) {
                    y = plot->Y.inf;
                    nyl++;
                }
                if (y >= plot->Y.sup) {
                    y = plot->Y.sup;
                    nyh++;
                }
                if (!plot->Y.force)
                    continue;
            }
            if (np == plot->maxpoints && np < n)
                cout << "reached maxpoints at " << s->x[i] << "\n";
            xp.push_back(x);
            yp.push_back(y);
            if (s->has_dy())
                dyp.push_back(s->dy[i]);
            np = xp.size();
            if (np > plot->maxpoints) // continue in large steps
                i += plot->maxpoints;
        }
        if (np == 0) {
            cout << "file " << k << " spec " << j << " all " << n << " points outside plot range:\n";
            if (nxl)
                cout << "  " << nxl << " points < xmin\n";
            if (nxh)
                cout << "  " << nxh << " points > xmax\n";
            if (nyl)
                cout << "  " << nyl << " points < ymin\n";
            if (nyh)
                cout << "  " << nyh << " points > ymax\n";
        } else {
            plot->add_spec(
                false, true, pstyle, xp, yp, dyp, fd->V[j]->z_str(), fd->xco.str_std(),
                fd->yco.str_std(), "data file " + S(k) + " spectrum " + S(j));
        }
        return np;
    }


//! Plot slice j of convolved curve file fc; return number of points plotted.

    int plot_curve_convolved(CPlot* plot, const COlc* fc, int k, int j, int cstyle)
    {
        vector<double> novec;
        int kconv, jconv;
        const CSpec* sconv;
        NCurveFile::get_conv(&sconv, &kconv, &jconv, k, j);
        vector<double> xp;
        for (int i = 0; i < sconv->size(); ++i) {
            double x = sconv->x[i];
            if (plot->X.contains(x))
                xp.push_back(x);
        }
        RObjVec cu = fc->eval_curve(xp, k, j, false);
        vector<double> xo, yo;
        for (int i = 0; i < xp.size(); ++i) {
            double yp = cu->to_r(i);
            if (!(plot->X.contains(xp[i]) && plot->Y.contains(yp)))
                continue;
            xo.push_back(xp[i]);
            yo.push_back(yp);
        }
        if (xo.size() == 0) {
            cout << "curve k=" << S(k) << ", j=" << S(j) << " has no points in plot window\n";
        }
        plot->add_spec(
            true, true, cstyle, xo, yo, novec, fc->V[j]->z_str(), fc->xco.str_std(), fc->yco.str_std(),
            "curve file " + S(k) + " spectrum " + S(j));
        return xo.size();
    }

//! Plot slice j of curve file fc, using grid from file kd.

    int plot_curve_to_grid(CPlot* plot, const COlc* fc, int k, int j, int cstyle)
    {
        if (fc->kd == -1)
            throw S("data reference not set, incompatible with cg+");
        const COld* fd = SMem::instance()->mem_get_D(fc->kd);
        if (!fd)
            throw S("data reference not found");
        if (j >= fd->nJ())
            throw S("number of spectra does not match");
        vector<double>* xc = &(fd->VS(j)->x);
        vector<double> novec, xp, yp;
        RObjVec cu = fc->eval_curve(*xc, k, j, false);
        for (int i = 0; i < xc->size(); ++i) {
            double yc = cu->to_r(i);
            if (plot->X.contains((*xc)[i]) && plot->Y.contains(yc)) {
                xp.push_back((*xc)[i]);
                yp.push_back(yc);
            }
        }
        plot->add_spec(
            true, true, cstyle, xp, yp, novec, fc->V[j]->z_str(), fc->xco.str_std(), fc->yco.str_std(),
            "curve file " + S(k) + " spectrum " + S(j));
        return xp.size();
    }


//! Plot slice j of non-convolved curve file fc on equidistant grid; return number of points plotted.

    int plot_curve_equidist(CPlot* plot, const COlc* fc, int k, int j, int cstyle)
    {
        vector<double> novec;
        // equidistant grid

        int npts = plot->equipoints;
        vector<double> xc, xp, yp;
        plot->X.set_xgrid(xc, npts);
        RObjVec cu = fc->eval_curve(xc, k, j, false);
        for (int i = 0; i < npts; ++i) {
            double yc = cu->to_r(i);
            if (plot->X.contains(xc[i]) && plot->Y.contains(yc)) {
                xp.push_back(xc[i]);
                yp.push_back(yc);
            }
        }
        plot->add_spec(
            true, true, cstyle, xp, yp, novec, fc->V[j]->z_str(), fc->xco.str_std(), fc->yco.str_std(),
            "curve file " + S(k) + " spectrum " + S(j));
        return xp.size();
    }

//! Plots slice j of non-convolved curve file fc, refining the grid where necessary;
//! return number of points plotted.

    int plot_curve_refine(CPlot* plot, const COlc* fc, int k, int j, int cstyle)
    {
        class CP2 {
        public:
            double x;
            double y;
            CP2 operator-(const CP2 other) {
                return { x-other.x, y-other.y }; };
            double operator*(const CP2 other) {
                return x*other.x + y*other.y; };
        };

        class CPt {
        public:
            double xd; // data point
            double yd;
            int code; // -1=nan 0=valid, 1=below, 2=above, 3=off_range
            CP2 P; // plot point
        };

        auto x2p = [&](const vector<double>& x) -> vector<CPt>
            {
                vector<double> y = PCAST<const CObjVecDbl>(fc->eval_curve(x, k, j, false))->v;
                vector<int> inRange = fc->eval_prange(x, k, j)->v;
                vector<CPt> ret(x.size());
                for (size_t i=0; i<x.size(); ++i) {
                    double xp = plot->X.pc(x[i]);
                    if (std::isnan(y[i]))
                        ret[i] = {x[i], y[i], 1, {xp, NAN}};
                    else if (y[i]<plot->Y.inf)
                        ret[i] = {x[i], y[i], 1, {xp, NAN}};
                    else if (y[i]>plot->Y.sup)
                        ret[i] = {x[i], y[i], 2, {xp, NAN}};
                    else if (!inRange[i])
                        ret[i] = {x[i], y[i], 3, {xp, NAN}};
                    else
                        ret[i] = {x[i], y[i], 0, {xp, plot->Y.pc(y[i])}};
                }
                return ret;
            };

        auto pt_order = [](const CPt a, const CPt b) -> bool
            { return a.xd<=b.xd; };

        auto nonlinearity = [](const vector<CPt>::iterator b) -> double
            {
                auto a = b-1;
                auto c = b+1;
                CP2 BA = a->P-b->P;
                CP2 AC = c->P-a->P;
                // squared  deviation from linearity:
                return BA*BA - pow(BA*AC,2)/(AC*AC);
            };

        // start with equidistant grid:
        vector<double> xn;
        plot->X.set_xgrid(xn, plot->equipoints);

        // refinement loop:
        vector<CPt> pp;
        for (int iref=0; iref<49 && pp.size() < plot->maxpoints; ++iref) {
            vector<CPt> pn = x2p(xn);
            pp = triv::merge_sorted(pp, pn, pt_order);
            xn.clear();

            for (auto b = pp.begin()+1; b<pp.end(); ++b) {
                auto a = b-1;
                if (a->code!=0 || b->code!=0) {
                    if (a->code!=b->code) {
                        xn.push_back( (b->xd + a->xd)/2 );
                    }
                } else { // valid data
                    auto c = b+1;
                    if (c<pp.end() && c->code==0) {
                        // squared  deviation from linearity:
                        if ( nonlinearity(b) > 1e-8 ) {
                            double dx = std::min(b->xd - a->xd, c->xd - b->xd)/2;
                            xn.push_back( b->xd - dx );
                            xn.push_back( b->xd + dx );
                        }
                    }
                }
            }
            if (!xn.size())
                break;
            xn = std::vector<double>( xn.begin(), std::unique(xn.begin(), xn.end()) );
        }

        // remove points in straight lines
        vector<CPt> po;
        size_t nout = 0;
        for (int iref=0; iref<49; ++iref) {
            po.clear();
            po.push_back( *pp.begin() );
            bool even = false;
            for (auto b = pp.begin()+1; b<pp.end()-1; ++b) {
                if (even || b->code!=(b-1)->code!=0 || b->code!=(b+1)->code!=0
                    || b->code==0 && nonlinearity(b)>1e-9)
                    po.push_back(*b);
                even = !even;
            }
            po.push_back( *(pp.end()-1) );
            if (po.size()==nout)
                break;
            nout = po.size();
            pp = po;
        }

        // divide into segments
        vector<vector<CPt>> segments;
        vector<CPt> seg;
        for (CPt p: po) {
            if (p.code==0)
                seg.push_back(p);
            else if (seg.size()) {
                segments.push_back(seg);
                seg.clear();
            }
        }
        if (seg.size())
            segments.push_back(seg);

        // plot segments
        size_t npts = 0;
        for (size_t is=0; is<segments.size(); ++is) {
            vector<double> xf, yf;
            for (CPt p: segments[is]) {
                xf.push_back(p.xd);
                yf.push_back(p.yd);
            }
            plot->add_spec(
                true, is==0, cstyle, xf, yf, vector<double>(), fc->V[j]->z_str(), fc->xco.str_std(),
                fc->yco.str_std(), "curve file " + S(k) + " spectrum " + S(j));
            npts += segments[is].size();
        }

        return npts;
    }

//**************************************************************************************************
//*  plot1D
//**************************************************************************************************

    //! Plot set of spectra, or add to open plot.

    void plot1D(class CPlot* plot, bool add, const string& mode)
    {
        FileIterator fiter(SFSel::instance()->sel());

        static vector<int> JSel;
        static int pstyle = 1, cstyle = 1;
        static CCoord xCo, yCo;

        // Get J selection.
        if (mode == "ask") {
            JSelAsk("Plot which spectra", JSel);
        } else {
            if (mode == "next") {
                triv::increment_indices(JSel, +1, SFSel::instance()->nJ_max());
            } else if (mode == "prev") {
                triv::increment_indices(JSel, -1, SFSel::instance()->nJ_max());
            } else
                throw "unexpected mode " + mode;
            if (!JSel.size())
                throw S("no spectra in range");
            cout << "plotting spectra " << triv::indices_to_s(JSel) << "\n";
        }

        // Prepare frame (ranges, labels).
        if (!add) {
            pstyle = 1;
            cstyle = 1;
            if (!plot->X.finite())
                determine_Xrange(plot, JSel);
            if (!plot->Y.finite())
                determine_Yrange(plot, JSel);
            xCo = CCoord();
            yCo = CCoord();
        }
        while (const COlo* f = fiter.next()) {
            xCo.merge(f->xco);
            yCo.merge(f->yco);
        }
        if (!add) {
            plot->start_frame1D("Frida version " VERSION, xCo.str_ps(), yCo.str_ps());
        } else
            plot->reopen_frame1D("Frida version " VERSION, xCo.str_ps(), yCo.str_ps());

        // Plot slices.
        fiter.reset();
        while (const COlo* f = fiter.next()) {
            int k = fiter.k();
            const COld* fd = dynamic_cast<const COld*>(f);
            const COlc* fc = dynamic_cast<const COlc*>(f);

            plot->doc_TxLine(f->name);
            for (int i = 0; i < f->lDoc.size(); i++)
                plot->doc_TxLine("  " + f->lDoc[i]);
            if (fc) {
                for (string line : fc->info_settings())
                    plot->doc_TxLine("  " + line);
                plot->doc_TxLine("  " + fc->info_table_header("e"));
            }

            // JSel.evaluate( 0, f->nJ()-1 );
            for (int iv = 0; iv < JSel.size(); ++iv) {
                int j = JSel[iv];

                if (fd) {
                    if (!plot_data(plot, fd, k, j, pstyle))
                        continue;
                    plot->doc_PtTxLine("  " + fd->VS(j)->info_line(), pstyle);
                    ++pstyle;

                } else if (fc) {
                    if (fc->kconv != -1) {
                        if (!plot_curve_convolved(plot, fc, k, j, cstyle))
                            continue;
                    } else if (fc->plot_to_grid) {
                        if (!plot_curve_to_grid(plot, fc, k, j, cstyle))
                            continue;
                    } else if (!plot->refine) {
                        if (!plot_curve_equidist(plot, fc, k, j, cstyle))
                            continue;
                    } else {
                        if (!plot_curve_refine(plot, fc, k, j, cstyle))
                            continue;
                    }
                    plot->doc_CvTxLine("  " + fc->VC(j)->info_line("e"), cstyle);
                    ++cstyle;
                }
            }
        }
        plot->show_specs();
    }

//**************************************************************************************************
//*  plot2D
//**************************************************************************************************

    //! Plot all spectra as 2D color plot

    void plot2D(class CPlot* plot)
    { // EMBEDDED DIALOG
        const COld* fd = SFSel::instance()->theOneD();

        size_t m = fd->nJ();
        if (m<2)
            throw "2d plot requires at least 2 slices";
        // determine bin limits
        vector<double> zval = fd->zvec(0)->to_rvec();
        vector<double> zlim = triv::histogram_limits(zval);
        vector<double> xlim = triv::histogram_limits(fd->VS(0)->x);
        double xinf = xlim.front();
        double xsup = xlim.back();
        double yinf, ysup;
        if (plot->Y.logflag)
            triv::getPosminMax(fd->VS(0)->y, yinf, ysup);
        else
            triv::getMinMax(fd->VS(0)->y, yinf, ysup);
        for (size_t j=1; j<m; ++j) {
            xlim = triv::histogram_limits(fd->VS(j)->x);
            xinf = std::min(xinf, xlim.front());
            xsup = std::max(xsup, xlim.back());
            double yyinf, yysup;
            if (plot->Y.logflag)
                triv::getPosminMax(fd->VS(j)->y, yyinf, yysup);
            else
                triv::getMinMax(fd->VS(j)->y, yyinf, yysup);
            yinf = std::min(yinf, yyinf);
            ysup = std::max(ysup, yysup);
        }
        if (!plot->X.finite())
            plot->X.set_limits(xinf, xsup);
        if (!plot->Z.finite())
            plot->Z.set_limits(zlim.front(), zlim.back());
        if (!(std::isfinite(yinf) && std::isfinite(ysup)))
            throw "No valid y values";
        if (!plot->Y.finite())
            plot->Y.set_limits(yinf-1e-4*std::abs(yinf), ysup+1e-4*std::abs(ysup));
        plot->X.check_single_precision();
        plot->Y.check_single_precision();
        plot->Z.check_single_precision();
        // plot
        plot->start_frame2D("Frida version " VERSION);
        bool recent_empty_line = false;
        for (size_t j=0; j<m; ++j) {
            const CSpec* s = fd->VS(j);
            vector<double> xlim = triv::histogram_limits(s->x);
            for (size_t i=0; i<s->size(); ++i) {
                double yout = s->y[i];
                if (  (plot->Y.force || plot->Y.contains(yout))
                    && plot->X.contains(xlim[i]) && plot->X.contains(xlim[i+1])
                    && plot->Z.contains(zlim[j]) && plot->Z.contains(zlim[j+1])) {
                    if (fabs(yout)<FLT_MIN)
                        yout = copysign(FLT_MIN, yout);
                    else if (fabs(yout)>FLT_MAX)
                        yout = copysign(FLT_MAX, yout);
                    plot->ps_line(
                        str(format("%13.7g wx %13.7g wx  %13.7g wy %13.7g wy  %13.7g wH t2d")
                            % xlim[i] % xlim[i+1] % zlim[j] % zlim[j+1] % yout ) );
                    recent_empty_line = false;
                }
            }
            if (!recent_empty_line) {
                plot->ps_line("");
                recent_empty_line = true;
            }
        }
        plot->close_frame2D(fd->xco.str_ps(), fd->ZCo[0].str_ps(), fd->yco.str_ps());
        // doclines
        plot->doc_TxLine(fd->name);
        for (int i = 0; i < fd->lDoc.size(); i++)
            plot->doc_TxLine("  " + fd->lDoc[i]);
        // plot -> file
        static string fname = "/tmp/frida-" + S(getpid()) + "-plowin" + S(plot->iPlot);
        plot->write_postscript(
            triv::wordexp_unique(fname), "overwrite", true, SVariRegistry::instance()->to_sMap());
        cout << "Plotted to " << fname << "\n";
        // Launch viewer - but only once per session.
        static string viewer = "";
        if (CNode::eval("$interactive")->to_i() && viewer=="") {
            viewer = CNode::eval("ps_viewer")->to_s();
            system((viewer+" "+fname+" >nul 2>nul &").c_str());
        }
    }

} // anonymous namespace

//**************************************************************************************************
//*  Main plot routine accessible from the main menu
//**************************************************************************************************

void NPlot::plot(class CPlot* plot, bool add, const string& mode)
{
    if (plot->p2D) {
        if (add)
            throw "Cannot 'add' to 2D plot";
        if (mode=="next" || mode=="prev")
            throw "Cannot scroll through 2D plot";
        ::plot2D(plot);
    } else
        ::plot1D(plot, add, mode);
}
