//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  slice.cpp
//! \brief CSlice handles z dependent data, is base class for CSpec, CCurve.

#include "defs.hpp"

#include <algorithm>

#include "../trivia/vector_ops.hpp"

#include "obj.hpp"
#include "ptr.hpp"
#include "slice.hpp"


//**************************************************************************************************
//*  class Slice
//**************************************************************************************************


//! Returns a vector of strings describing the z entries.

vector<string> CSlice::z_str() const
{
    vector<string> ret;
    for (RObj zi : z)
        ret.push_back(zi->to_s());
    return ret;
}


//**************************************************************************************************
//*  class CSpec
//**************************************************************************************************


//! Append given data point.

void CSpec::push_xy(double _x, double _y)
{
    x.push_back(_x);
    y.push_back(_y);
}


//! Append given data point, with error.

void CSpec::push_xyd(double _x, double _y, double _dy)
{
    if (y.size() != dy.size())
        throw S("cannot push_xyd since y and dy have different size");
    x.push_back(_x);
    y.push_back(_y);
    dy.push_back(_dy);
}


//! Remove data points according to given indices.

void CSpec::remove(int from, int to)
{
    x.erase(x.begin() + from, x.begin() + to);
    y.erase(y.begin() + from, y.begin() + to);
    if (has_dy())
        dy.erase(dy.begin() + from, dy.begin() + to);
}


//! Reset to initial state.

void CSpec::clear()
{
    x.clear();
    y.clear();
    dy.clear();
    z.clear();
}


//! Resize x,y,dy.

void CSpec::resize(int n, bool with_dy)
{
    x.resize(n);
    y.resize(n);
    if (with_dy)
        dy.resize(n);
}


//! Return number of data points.

int CSpec::size() const
{
    if (x.size() != y.size())
        throw "BUG detected by spec::size: #x [" + S(x.size()) + "] <> #y [" + S(y.size()) + "]\n";
    if (dy.size() != y.size() && has_dy())
        throw "BUG detected by spec::size: #y [" + S(y.size()) + "] <> #dy [" + S(dy.size())
            + "]\n";
    return y.size();
}


//! For each xx[i], set yy[i] by interpolating in this.x -> this.y.

void CSpec::intpol(const vector<double>& xx, vector<double>* yy, vector<double>* dd) const
{
    int n = size();
    if (n < 2)
        throw "invalid interpolation request: old grid has size " + S(n);
    int nn = xx.size();
    if (nn < 1)
        throw S("invalid interpolation request: new grid is empty");
    yy->resize(nn, 0.);
    if (dd)
        dd->resize(nn, 0.);

    if (!triv::is_ascending(x))
        throw S("cannot interpolate: old grid not sorted");
    if (!triv::is_ascending(xx))
        throw S("cannot interpolate: new grid not sorted");

    if (n == nn) { // test for equality
        double span = x[n - 1] - x[0];
        for (int i = 0; i < n; ++i) {
            if (fabs(x[i] - xx[i]) > 1e-14 * span)
                goto scales_differ;
        }
        for (int i = 0; i < n; ++i) {
            (*yy)[i] = y[i];
            if (dd)
                (*dd)[i] = dy[i];
        }
        return;
    }
scales_differ:

    /*
    if( strict && ( xx[0]<x[0] || xx[nn-1]>x[n-1] ) )
        throw "interpolation to "+S(xx[0])+" .. "+S(xx[nn-1])+
            " required, while available data only cover "+S(x[0])+" .. "+S(x[n-1]);
    */

    for (int ii = 0; ii < nn; ++ii) {
        (*yy)[ii] = 0;
        if (dd)
            (*dd)[ii] = 0;
    }
    int i = 0, ii = 0;
    while (xx[ii] < x[0]) {
        ++ii;
        if (ii >= nn)
            return;
    }
    while (i < n - 1 && xx[ii] < x[n - 1]) {
        while (xx[ii] > x[i + 1] && i < n - 1)
            ++i;
        if (i >= n - 1)
            break;
        if (ii >= nn)
            return;
        (*yy)[ii] = ((x[i + 1] - xx[ii]) * y[i] + (xx[ii] - x[i]) * y[i + 1]) / (x[i + 1] - x[i]);
        // std::cout << "DEBUG INTPOL i=" << i << " ii=" << ii << " x=" << x[i] << ".." << x[i+1] <<
        //    " xx=" << xx[ii] << " y=" << y[i] << ".." << y[i+1] << " -> " << (*yy)[ii] << "\n";
        if (dd) {
            double genus = ((x[i + 1] - xx[ii]) * SQR(dy[i]) + (xx[ii] - x[i]) * SQR(dy[i + 1]))
                / (x[i + 1] - x[i]);
            if (genus < 0)
                throw "interpolation failed: variance<0";
            (*dd)[ii] = sqrt(genus);
        }
        ++ii;
    }
}

//! Compute norm (trapezoidal sum?).

double CSpec::norm(string info) const
{
    int nv = size();
    if (nv < 2)
        throw "cannot compute norm of " + info + ": only " + S(nv) + " points";
    // Calculate normalization integral.
    // At this occasion, assert that x is ordered,
    double ret = 0;
    // special treatment for first and last step:
    double dx = x[1] - x[0];
    if (dx <= 0)
        throw "cannot compute norm of " + info + ": not ordered in x";
    ret += y[0] * dx;
    dx = x[nv - 1] - x[nv - 2];
    if (dx <= 0)
        throw "cannot compute norm of " + info + ": not ordered in x";
    ret += y[nv - 1] * dx;
    // the other steps:
    for (int i = 1; i < nv - 1; ++i) {
        dx = (x[i + 1] - x[i - 1]) / 2;
        if (dx <= 0)
            throw "cannot compute norm of " + info + ": not ordered in x";
        ret += y[i] * dx;
    }
    // check result:
    if (ret == 0)
        throw "norm of " + info + " is zero";
    else if (ret <= 0)
        throw "norm of " + info + " is negativ";
    return ret;
}

//! Is x equidistant with a given step?

bool CSpec::has_step(double step) const
{
    if (!step)
        return false;
    for (int i = 0; i < size() - 1; ++i) {
        if (fabs(x[i + 1] - x[i] - step) > 1e-12 * step)
            return false;
    }
    return true;
}

//! Return line containing z values.

string CSpec::info_line() const
{
    const int LEN = 15;
    char wrd[LEN];
    string out = "";
    for (int iz = 0; iz < z.size(); iz++) {
        snprintf(wrd, LEN, "%-13.6g", z[iz]->to_r());
        out += wrd;
        if (iz < z.size() - 1)
            out += " ";
    }
    return out;
}

//! Returns true if all data points have nonzero dy.

bool CSpec::dy_are_all_nonzero() const
{
    if (!has_dy())
        return false;
    return std::all_of(dy.cbegin(), dy.cend(), [](double dv) { return dv > 0; });
}

PSpec CSpec::binned(const vector<int>& breaks, const bool sampling_error) const
{
    PSpec sout(new CSpec());
    sout->z = z;
    for (int iv = 0; iv < breaks.size() - 1; ++iv) {
        int igi = breaks[iv];
        int igf = breaks[iv + 1];
        double xg = 0, yg = 0;
        int ng = igf - igi;
        for (int i = igi; i < igf; ++i) {
            xg += x[i];
            yg += y[i];
        }
        xg /= ng;
        yg /= ng;
        double vg_src = 0, vg_grp = 0;
        for (int i = igi; i < igf; ++i) {
            if (has_dy())
                vg_src += SQR(dy[i]);
            if (sampling_error)
                vg_grp += SQR(y[i] - yg);
        }
        double vg = vg_src / ng / ng;
        if (sampling_error && ng > 1)
            vg += vg_grp / (ng - 1);
        sout->push_xyd(xg, yg, sqrt(vg));
    }
    return sout;
}

//**************************************************************************************************
//*  class CCurve
//**************************************************************************************************


//! Initialize curve. Used when creating or loading a curve file.

void CCurve::clear_fit_outcome()
{
    fitOutcome = 0;
    fitChi2 = 0;
    fitR2 = 0;
}

//! Asserts that ParAttr has a valid value.

void CCurve::assert_attr() const
{
    for (int ip = 0; ip < P.size(); ip++)
        if (!(ParAttr[ip] == 'u' || ParAttr[ip] == 'x' || ParAttr[ip] == 'g'))
            throw "BUG: ParAttr has invalid value '" + S(ParAttr[ip]) + "'";
}

//! Returns true if there is at least one global parameter.

bool CCurve::has_global_par() const
{
    for (char c : ParAttr)
        if (c == 'g')
            return true;
    return false;
}

//! Returns line containing z and p values.

string CCurve::info_line(const string& which) const
{
    const int LEN = 25;
    char wrd[LEN];
    string out = "";
    for (int iz = 0; iz < z.size(); iz++) {
        snprintf(wrd, LEN, "%-13.6g", z[iz]->to_r());
        out += wrd;
        out += " ";
    }
    assert_attr();
    for (int ip = 0; ip < P.size(); ip++) {
        out += (ParAttr[ip] == 'u') ? ' ' : ParAttr[ip];
        snprintf(wrd, LEN, "%-13.6g", P[ip]);
        out += wrd;
        if (which=="e") {
            snprintf(wrd, LEN, " %-11.4g", dP(ip));
            out += wrd;
        }
        if (ip < P.size() - 1)
            out += " ";
    }
    snprintf(wrd, LEN, "%-2i ", fitOutcome);
    out += wrd;
    snprintf(wrd, LEN, "%-22.16g", fitChi2);
    out += wrd;
    snprintf(wrd, LEN, "%-22.16g", 1 - fitR2);
    out += wrd;
    return out;
}

//! Returns parameter uncertainty, estimated from diagonal element of the covariance matrix.

double CCurve::dP(int ip) const
{
    return sqrt(PCovar[ip*P.size()+ip]);
}

//! Prints covariance matrix.
void CCurve::print_par_covar(const string& which) const
{
    int n = P.size();
    for (int j=0; j < n; j++) {
        for (int i=0; i < n; i++) {
            if      (which=="a")
                printf("%12.5f ", PCovar[j*n+i]);
            else if (which=="r")
                printf("%12.5f ", PCovar[j*n+i]/P[i]/P[j]);
            else if (which=="p")
                printf("%12.5f ", PCovar[j*n+i]/sqrt(PCovar[j*n+j]*PCovar[i*n+i]));
            else
                throw "Invalid command cpc"+which;
        }
        printf("\n");
    }
}
