//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, C++ rewrite 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file toplevel.cpp
//! \brief Execution context CFrida.

#include "defs.hpp"

#include <fstream>
#include <gsl/gsl_errno.h>

#include "../plot/plowin.hpp"
#include "../readplus/macro.hpp"
#include "../readplus/readln.hpp"
#include "../trivia/vector_ops.hpp"

#include "commands.hpp"
#include "expr.hpp"
#include "fsel.hpp"
#include "func.hpp"
#include "loop.hpp"
#include "mem.hpp"
#include "obj.hpp"
#include "olf.hpp"
#include "reduce_curv.hpp"
#include "reduce_spec.hpp"
#include "toplevel.hpp"
#include "variables.hpp"
#include "xax_lex.hpp"

// Declarations.
void fbase_initialize(); // implemented in fbase.cpp; no .hpp file
void fassign_initialize(); // implemented in fassign.cpp; no .hpp file
void fstring_initialize(); // implemented in fstring.cpp; no .hpp file


//**************************************************************************************************
//*  Internals
//**************************************************************************************************

//! Error handler for C code from the GNU Scientific Library.

void my_gsl_error_handler(const char* reason, const char* file, int line, int gsl_errno)
{
    cout << "GSL error " << gsl_errno << ": " << reason << " in " << file << ", line " << line
         << "\n";
    // throw does not work (probably because GSL is in C, not in C++)
}


//! At exit, currently nothing to do.

void atexit_handler() {}


//! Executes one Frida command.

void CFrida::execute_cmd(const string& cmdline)
{
    // Empty input, comment lines.
    if (cmdline == "" || cmdline[0] == '#') {
        return;
    }

    string::size_type j = cmdline.find_first_of(" \t");
    if (j != string::npos)
        NMacro::push_remainder(cmdline.substr(j));
    string cmd = cmdline.substr(0, j);

    // function keys (TEMPORARY WORKAROUND)
    if (cmd.substr(0, 2) == "FK" && cmd.find("=") == string::npos) {
        RObj ret = SVariRegistry::instance()->find(cmd);
        if (!ret)
            throw "uninitialized function key " + cmd;
        cmd = ret->to_s();
    }

    // Is it a command?
    clock_cmd_i = clock();
    if (cmd == "t") {
        double time_total = ((double)(clock() - clock_start)) / CLOCKS_PER_SEC;
        printf("total frida CPU time now %g\n", time_total);
        printf("last command took %g\n", time_cmd);
        return;
    } else if (frida_command(cmd)) {
        ;
    } else {
        RNode T = user_xaxparse(cmd.c_str());
        if (cmd[0] != '(' && T->looks_like_indices()) {
            RObjVecInt v = T->to_index_list(SMem::instance()->mem_size());
            if (!v)
                throw S("not a valid index list");
            SFSel::instance()->set_selection(v);
        } else {
            // Pocket calculator
            if (T->has_dummy())
                throw S("dummy argument t not allowed outside curve definitions");
            if (T->k_dependent()) {
                FileIterator fiter(SFSel::instance()->sel());
                while ((fiter.next())) {
                    int k = fiter.k();
                    if (fiter.size() > 1)
                        cout << "f" << k << ":";
                    cout << T->tree_point_val(k)->to_s() << "\n";
                }
            } else {
                RObj ret = T->tree_point_val();
                if (ret->to_s() != "__nullptr__")
                    cout << ret->to_s() << "\n";
            }
        }
    }
    time_cmd = ((double)(clock() - clock_cmd_i)) / CLOCKS_PER_SEC;
}


//**************************************************************************************************
//*  API
//**************************************************************************************************

//! Prepare interactive Frida session.

CFrida::CFrida()
{
    int line = 0;
    try {
        // Initializations:
        NReadln::initialize(); // command-line dialog (readline and history)
        line = 1;
        SPloWin::initialize(); // plot windows
        line = 2;
        gsl_set_error_handler(my_gsl_error_handler);
        line = 3;
        fbase_initialize();
        line = 4;
        fassign_initialize();
        line = 5;
        fstring_initialize();
        line = 6;
        SGeniRegistry::initialize();
        line = 7;
        SCvinRegistry::initialize();
        line = 8;
        if (std::atexit(atexit_handler)) {
            cout << "BUG: atexit registration failed\n";
            exit(EXIT_FAILURE);
        }
        // time monitor:
        line = 9;
        clock_start = clock();
        clock_cmd_i = clock();
        time_cmd = 0;
    } catch (string& ex) {
        cout << "CFrida initialization failed in line " << line << "\n  " << ex << "\n";
        exit(1);
    } catch (const char* ex) {
        cout << "CFrida initialization failed in line " << line << "\n  " << ex << "\n";
        exit(1);
    } catch (std::exception& ex) {
        cout << "CFrida initialization failed in line " << line << "\n  " << ex.what() << "\n";
        exit(1);
    } catch (...) {
        cout << "CFrida initialization failed in line " << line
             << " with unforeseen exception type\n";
        exit(1);
    }
}


//! Prompts user for a command, and returns it.

void CFrida::interactive()
{
    execute_cmd("ev $interactive 1");
    while (1) {
        string cmdline;
        try {
            cmdline = NMacro::readln(triv::indices_to_s(SFSel::instance()->sel()) + " > ");
            if (cmdline=="EOF") {
                cout << "\n";
                exit(0);
            }
            execute_cmd(cmdline);
        } catch (string& ex) {
            cout << "'" << cmdline << "' failed:\n  " << ex << "\n";
            NMacro::clear();
        } catch (const char* ex) {
            cout << "'" << cmdline << "' failed:\n  " << ex << "\n";
            NMacro::clear();
        } catch (std::exception& ex) {
            cout << "BUG: '" << cmdline << "' failed with unforeseen std exception:\n  "
                 << ex.what() << "\n";
            NMacro::clear();
        } catch (...) {
            cout << "BUG: '" << cmdline << "' failed with unforeseen exception\n";
            NMacro::clear();
        }
    }
}


//! Reads a script file and executes it line by line.

void CFrida::execute_file(const string& fnam, bool batchMode)
{
    execute_cmd("ev $interactive 0");
    int lineno = 0;
    std::ifstream F(fnam);
    if (!F.is_open()) {
        cout << "File " << fnam << " not found\n";
        return;
    }
    try {
        cout << "executing " << fnam << "\n";
        while (true) {
            string cmdline = NMacro::readln("", &F, &lineno);
            if (cmdline == "EOF")
                break;
            try {
                execute_cmd(cmdline);
            } catch (string& ex) {
                string msg = "'" + cmdline + "':\n  " + ex;
                if (batchMode) {
                    std::cerr << msg << "\n";
                    exit(1);
                }
                throw msg;
            } catch (const char* ex) {
                string msg = "'" + cmdline + "':\n  " + ex;
                if (batchMode) {
                    std::cerr << msg << "\n";
                    exit(1);
                }
                throw msg;
            } catch (...) {
                string msg = "'" + cmdline + "':\n  unforeseen exception type";
                if (batchMode) {
                    std::cerr << msg << "\n";
                    exit(1);
                }
                throw msg;
            }
        }
    } catch (string& ex) {
        cout << "Error in script " << fnam << ", line " << lineno << ":\n  " << ex << "\n";
    } catch (const char* ex) {
        cout << "BUG: char* error in script " << fnam << ", line " << lineno << ":\n  " << ex
             << "\n";
    } catch (std::exception& ex) {
        cout << "BUG: stdlib exception in script " << fnam << ", line " << lineno << ":\n  "
             << ex.what() << "\n";
    } catch (...) {
        cout << "BUG: script " << fnam << " failed with unforeseen exception\n";
    }
}
